""" fprime.fpp.cli: FPP command line targets

Processing and command line functions for FPP tools wrappers in fprime-util.

@mstarch
"""
import argparse
from typing import Dict, List, Tuple, Callable
from fprime.fpp.common import FppUtility


def run_fpp_check(
    build: "Build",
    parsed: argparse.Namespace,
    _: Dict[str, str],
    __: Dict[str, str],
    ___: List[str],
):
    """Run fpp check application

    Handles the fpp-check endpoint by running the utility fpp-check.

    Args:
        build: build directory output
        parsed: parsed input arguments
        _: unused cmake_args
        __: unused make_args
        ___: unused pass-through arguments
    """
    FppUtility("fpp-check").execute(
        build,
        parsed.path,
        args=({}, ["-u", parsed.unconnected] if parsed.unconnected else []),
    )


def add_fpp_parsers(
    subparsers, common: argparse.ArgumentParser
) -> Tuple[Dict[str, Callable], Dict[str, argparse.ArgumentParser]]:
    """Sets up the fpp command line parsers

    Creates command line parsers for fpp commands and associates these commands to processing functions for those fpp
    commands.

    Args:
        subparsers: subparsers to add to
        common: common parser for all fprime-util commands

    Returns:
        Tuple of dictionary mapping command name to processor, and command to parser
    """
    check_parser = subparsers.add_parser(
        "fpp-check", help="Runs fpp-check utility", parents=[common], add_help=False
    )
    check_parser.add_argument(
        "-u", "--unconnected", default=None, help="write unconnected ports to file"
    )
    return {"fpp-check": run_fpp_check}, {"fpp-check": check_parser}



"""
Created on Dec 18, 2014

@author: tcanham

"""

# Exception classes for all types
from fprime.common.error import FprimeException


class TypeException(FprimeException):
    """An exception in our python types"""

    def __init__(self, val):
        super().__init__(val)
        self.except_msg = val

    def getMsg(self):
        """Gets the exception message"""
        return self.except_msg


class AbstractMethodException(TypeException):
    """Not implemented exception under another name"""

    def __init__(self, val):
        super().__init__(f"{str(val)} must be implemented since it is abstract!")


class TypeRangeException(TypeException):
    """Value is out of range"""

    def __init__(self, val):
        super().__init__(f"Value {str(val)} out of range!")


class StringSizeException(TypeException):
    """String size is to large for defined type"""

    def __init__(self, size, max_size):
        super().__init__(f"String size {str(size)} is greater than {str(max_size)}!")


class TypeMismatchException(TypeException):
    """Wrong type found exception"""

    def __init__(self, expected_type, actual_type):
        super().__init__(f"Type {expected_type} expected, type {actual_type} found!")


class ArrayLengthException(TypeException):
    """Array length mismatched"""

    def __init__(self, arr_type, expected_len, actual_len):
        super().__init__(
            f"Array type {arr_type} is of length {expected_len}, actual length {actual_len} found!"
        )


class EnumMismatchException(TypeException):
    """Enum member not defined"""

    def __init__(self, enum, bad_member):
        super().__init__(f"Invalid enum member {bad_member} set in {enum} enum!")


class MissingMemberException(TypeException):
    """Member was not defined on type"""

    def __init__(self, field):
        super().__init__(f"Value does not define required field: {field}")


class IncorrectMembersException(TypeException):
    """Members incorrectly defined on type (too many, too few)"""

    def __init__(self, fields):
        super().__init__(f"Value does not define required fields: {fields}")


class DeserializeException(TypeException):
    """Exception during deserialization"""

    pass


class ArgNotFoundException(TypeException):
    """Argument not found exception"""

    def __init__(self, message):
        super().__init__(f"Arg {message} not found!")


class NotInitializedException(TypeException):
    """Did not initialize types"""

    def __init__(self, message):
        super().__init__(f"Instance {message} not initialized!")


class NotOverriddenException(TypeException):
    """Not implemented exception by another name"""

    def __init__(self, message):
        super().__init__(
            f"Required base class method not overwritten in type {message}!"
        )


class ArgLengthMismatchException(TypeException):
    """Mismatch in args lengths vs definition of args"""

    def __init__(self, arg_length_actual, arg_length_given):
        super().__init__(
            "%d args provided, but command expects %d args!"
            % (arg_length_given, arg_length_actual)
        )


class CompoundTypeLengthMismatchException(TypeException):
    """Compound type fields mismatch"""

    def __init__(self, field_length_actual, field_length_given):
        super().__init__(
            "%d fields provided, but compound type expects %d fields!"
            % (field_length_given, field_length_actual)
        )



""" fprime.util.cli: general CLI handling

Sets up parsers and processors for general CLI targets under fprime-util that do not fit elsewhere. Includes parsers
such as: hast-to-file, info, and others.

@author mstarch
"""
import argparse
import sys

from pathlib import Path
from typing import Dict, Callable
from fprime.fbuild.builder import Build, InvalidBuildCacheException
from fprime.fbuild.interaction import new_component, new_port


def print_info(
    base: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str], ___
):
    """Builds and prints the informational output block

    Handles the info command using the given build as a template to discover other build types. This will load the
    information and prints it to the screen.

    Args:
        base: base build used to find other builds including information
        parsed: parsed command namespace
        _: unused cmake arguments
        __: unused make arguments
        ___: unused pass through arguments
    """
    # Roll up targets for more concise display. Dictionaries are used as their keys become a set and as of Python 3.7
    # the native dictionary maintains ordering. Thus it acts as an "ordered set" for a nice predictable display order
    # derived from the order of target definitions.
    build_infos = {}
    local_generic_targets = {}
    global_generic_targets = {}
    # Loop through available builds and harvest targets
    for build in Build.get_build_list(base, parsed.build_cache):
        build_info = build.get_build_info(Path(parsed.path))
        # Target list
        local_targets = {
            f"'{target}'": "" for target in build_info.get("local_targets", [])
        }
        global_targets = {
            f"'{target}'": "" for target in build_info.get("global_targets", [])
        }
        build_artifacts = (
            build_info.get("auto_location")
            if build_info.get("auto_location") is not None
            else "N/A",
            build_info.get("build_dir", "Unknown"),
        )
        local_generic_targets.update(local_targets)
        global_generic_targets.update(global_targets)
        build_infos[build.build_type] = build_artifacts

    # Print out directory and deployment target sections
    if local_generic_targets.keys() or global_generic_targets.keys():
        print("[INFO] fprime build information:")
        print(
            f"    Available directory targets: {' '.join(local_generic_targets.keys())}"
        )
        print()
        print(
            f"    Available global targets: {' '.join(global_generic_targets.keys())}"
        )
        print("  ----------------------------------------------------------")
    # Artifact locations come afterwards
    for (
        build_type,
        (build_artifact_location, global_build_cache),
    ) in build_infos.items():
        print(
            f"    {build_type.get_cmake_build_type()} build cache module directory: {build_artifact_location}"
        )
        print(
            f"    {build_type.get_cmake_build_type()} build cache: {global_build_cache}"
        )
    print()


def hash_to_file(
    build: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str], ___
):
    """Processes hash-to-file to locate file

    Args:
        build: build to search for hashes.txt
        parsed: parsed arguments for needed for parsed.hash
        _: unused cmake arguments
        __: unused make arguments
        ___: unused pass through arguments
    """
    lines = build.find_hashed_file(parsed.hash)
    if not lines:
        raise InvalidBuildCacheException(
            f"Hash 0x{parsed.hash:x} not found. Do you need '--ut' for a unittest run?"
        )
    print("[INFO] File(s) associated with hash 0x{:x}".format(parsed.hash))
    for line in lines:
        print("   ", line, end="")


def template(
    build: Build, parsed: argparse.Namespace, _: Dict[str, str], __: Dict[str, str]
):
    """Processes new command

    Args:
        build: build used to inform new component and new port calls
        parsed: parsed arguments
        _: unused cmake arguments
        __: unused make arguments
    """
    if parsed.component and not parsed.port:
        return new_component(build.deployment, parsed.platform, parsed.verbose, build)
    elif parsed.port and not parsed.component:
        return new_port(build.deployment, build)
    print("[ERROR] Use --component or --port, not both.", file=sys.stderr)


def add_special_parsers(
    subparsers, common: argparse.ArgumentParser, help_text: "HelpText"
) -> Dict[str, Callable]:
    """Adds in CLI parsers for other commands

    Args:
        subparsers: subparsers used to create new CLI subparsers
        common: common parent parser

    Returns:
        Dictionary associating special command name to callable used to process it
    """
    # Add a search for hash function
    hash_parser = subparsers.add_parser(
        "hash-to-file",
        description=help_text.long("hash-to-file"),
        help=help_text.short("hash-to-file"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    hash_parser.add_argument(
        "hash",
        type=lambda x: int(x, 0),
        help="Assert hash value to convert to filename",
    )

    # Add a search for hash function
    subparsers.add_parser(
        "info",
        description=help_text.long("info"),
        help=help_text.short("info"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # New functionality
    new_parser = subparsers.add_parser(
        "new",
        description=help_text.long("new"),
        help=help_text.short("new"),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    new_parser.add_argument(
        "--component",
        default=False,
        action="store_true",
        help="Tells the new command to generate a component",
    )
    new_parser.add_argument(
        "--port",
        default=False,
        action="store_true",
        help="Tells the new command to generate a port",
    )
    return {"hash-to-file": hash_to_file, "info": print_info, "new": template}



"""
Provided constants for use within the fprime system.
"""
# pylint: disable=W0105
"""
In order to transmit data in a serialized format, string objects need to be encoded. Otherwise,
it is unclear how the characters are translated into raw bytes on the wire. This value should be
consistent with the encoding used on the flight software that is being communicated with.

Traditional C/C++ strings typically use "ascii" encoding. Hence being used here.  However, should
F prime be updated to use some other encoding, this value may be changed.
"""
DATA_ENCODING = "utf-8"



""" fprime.fbuild.target_definitions: targets definitions for fprime-util

Defines all the targets for fprime-util. Each target is a singleton that is registered into the list of all targets and
as such, each target need only be instantiated but need not be assigned to anything.

"""
from .types import BuildType
from .target import TargetScope, BuildSystemTarget
from .gcovr import GcovrTarget

#### "build" targets for components, deployments, unittests for both normal and testing builds ####
BuildSystemTarget(
    "",
    mnemonic="build",
    desc="Build components, ports, and deployments",
    scope=TargetScope.BOTH,
)
BuildSystemTarget(
    "ut_exe",
    mnemonic="build",
    desc="Build unittests",
    scope=TargetScope.LOCAL,
    flags={"ut"},
    build_type=BuildType.BUILD_TESTING,
)
BuildSystemTarget(
    "all",
    mnemonic="build",
    desc="Build components, ports, UTs, and deployments for unittest build",
    scope=TargetScope.GLOBAL,
    flags={"all", "ut"},
    build_type=BuildType.BUILD_TESTING,
)


#### Implementation targets ####
BuildSystemTarget(
    "impl",
    mnemonic="impl",
    desc="Generate implementation template files",
    scope=TargetScope.LOCAL,
)
BuildSystemTarget(
    "testimpl",
    mnemonic="impl",
    desc="Generate unit test files",
    flags={"ut"},
    build_type=BuildType.BUILD_TESTING,
    scope=TargetScope.LOCAL,
)


#### Check targets ####
check = BuildSystemTarget(
    "check",
    mnemonic="check",
    desc="Build and run unittests",
    build_type=BuildType.BUILD_TESTING,
    scope=TargetScope.BOTH,
)
BuildSystemTarget(
    "check_leak",
    mnemonic="check",
    desc="Build and run unittests with memory leak checking",
    flags={"leak"},
    build_type=BuildType.BUILD_TESTING,
    scope=TargetScope.BOTH,
)
GcovrTarget(
    check,
    mnemonic="check",
    desc="Build run and calculate coverage of unittests",
    build_type=BuildType.BUILD_TESTING,
    scope=TargetScope.BOTH,
    flags={"coverage"},
)



from fprime.fbuild.interaction import is_valid_name

# Check to ensure Component Name is valid
def verify_inputs(component_name, commands, events, telemetry, parameters):
    if is_valid_name(component_name) != "valid":
        raise ValueError(
            "Unacceptable component name. Do not use spaces or special characters"
        )
    if commands == "no" and events == "no" and telemetry == "no" and parameters == "no":
        raise ValueError(
            "[ERROR] You must select at least one of the following options to have in your component: commands, events, telemetry, parameters"
        )


verify_inputs(
    "{{ cookiecutter.component_name }}",
    "{{ cookiecutter.commands }}",
    "{{ cookiecutter.events }}",
    "{{ cookiecutter.telemetry }}",
    "{{ cookiecutter.parameters }}",
)



"""
fprime.fbuild.settings:

An implementation used to pull settings into the fprime build.  This version uses INI files in order to load the
settings from the settings.default file that is part of the F prime deployment directory.

@author mstarch
"""
import os
import configparser
from typing import Dict, List
from pathlib import Path


class IniSettings:
    """Class to load settings from INI files"""

    DEF_FILE = "settings.ini"
    SET_ENV = "FPRIME_SETTINGS_FILE"

    @staticmethod
    def find_fprime(cwd: Path) -> Path:
        """
        Finds F prime by recursing parent to parent until a matching directory is found.
        """
        needle = Path("cmake/FPrime.cmake")
        path = cwd.resolve()
        while path != path.parent:
            if Path(path, needle).is_file():
                return path
            path = path.parent
        raise FprimeLocationUnknownException(
            "Please set 'framework_path' in [fprime] section in 'settings.ini"
        )

    @staticmethod
    def read_safe_path(
        parser: configparser.ConfigParser,
        section: str,
        key: str,
        ini_file: Path,
        exists: bool = True,
    ) -> List[str]:
        """
        Reads path(s), safely, from the config parser.  Validates the path(s) exists or raises an exception. Paths are
        separated by ':'.  This will also expand relative paths relative to the settings file.

        :param parser: parser to read from
        :param key: key to read from
        :param ini_file: ini_file path for clean errors
        :return: path, validated
        """
        base_dir = os.path.dirname(ini_file)
        all_paths = parser.get(section, key, fallback="").split(":")
        expanded = []
        for path in all_paths:
            if not path:
                continue
            full_path = os.path.abspath(os.path.normpath(os.path.join(base_dir, path)))
            if exists and not os.path.exists(full_path):
                raise FprimeSettingsException(
                    f"Nonexistent path '{path}' found in section '{section}' option '{key}' of file '{ini_file}'"
                )
            expanded.append(full_path)
        return expanded

    @staticmethod
    def load(settings_file: Path, platform: str = "fprime"):
        """
        Load settings from specified file or from specified build directory. Either a specific file or the build
        directory must be not None.

        :param settings_file: file to load settings from (in INI format). Must be specified if build_dir is not.
        :param platform: platform to read platform specific settings
        :return: a dictionary of needed settings
        """
        settings_file = (
            Path.cwd() / IniSettings.DEF_FILE
            if settings_file is None
            else settings_file
        ).resolve()
        dfl_install_dest = Path(settings_file.parent, "build-artifacts")

        # Check file existence if specified
        if not os.path.exists(settings_file):
            print(f"[WARNING] Failed to find settings file: {settings_file}")
            fprime_location = IniSettings.find_fprime(settings_file.parent)
            return {"framework_path": fprime_location, "install_dest": dfl_install_dest}
        confparse = configparser.ConfigParser()
        confparse.read(settings_file)
        # Search through F prime locations
        fprime_location = IniSettings.read_safe_path(
            confparse, "fprime", "framework_path", settings_file
        )
        fprime_location = (
            Path(fprime_location[0])
            if fprime_location
            else IniSettings.find_fprime(settings_file.parent)
        )
        # Read project root if it is available
        proj_root = IniSettings.read_safe_path(
            confparse, "fprime", "project_root", settings_file
        )
        proj_root = proj_root[0] if proj_root else None
        # Read ac constants if it is available
        ac_consts = IniSettings.read_safe_path(
            confparse, platform, "ac_constants", settings_file
        )
        ac_consts = ac_consts[0] if ac_consts else None
        # Read include constants if it is available
        config_dir = IniSettings.read_safe_path(
            confparse, platform, "config_directory", settings_file
        )
        config_dir = config_dir[0] if config_dir else None

        install_dest = IniSettings.read_safe_path(
            confparse, platform, "install_dest", settings_file, False
        )

        install_dest = Path(install_dest[0]) if install_dest else dfl_install_dest

        # Read separate environment file if necessary
        env_file = IniSettings.read_safe_path(
            confparse, platform, "environment_file", settings_file
        )
        env_file = env_file[0] if env_file else settings_file
        libraries = IniSettings.read_safe_path(
            confparse, platform, "library_locations", settings_file
        )
        environment = IniSettings.load_environment(env_file)

        settings = {
            "settings_file": settings_file,
            "framework_path": fprime_location,
            "library_locations": libraries,
            "default_toolchain": confparse.get(
                "fprime", "default_toolchain", fallback="native"
            ),
            "default_ut_toolchain": confparse.get(
                "fprime", "default_ut_toolchain", fallback="native"
            ),
            "install_dest": install_dest,
            "environment_file": env_file,
            "environment": environment,
            "component_cookiecutter": confparse.get(
                "fprime", "component_cookiecutter", fallback="default"
            ),
        }
        # Set the project root
        if proj_root is not None:
            settings["project_root"] = proj_root
        # Set AC constants if available
        if ac_consts is not None:
            settings["ac_constants"] = ac_consts
        # Set the config dir
        if config_dir is not None:
            settings["config_dir"] = config_dir
        return settings

    @staticmethod
    def load_environment(env_file):
        """
        Load the environment from the given parser.

        :param env_file: load environment from this file
        :return: environment dictionary
        """
        parser = configparser.ConfigParser()
        parser.optionxform = str
        parser.read(env_file)
        env_dict = {}
        try:
            for key, value in parser.items("environment"):
                env_dict[key] = value
        except configparser.NoSectionError:
            pass  # Ignore missing environment
        return env_dict


class FprimeLocationUnknownException(Exception):
    """Fprime location could not be determined"""


class FprimeSettingsException(Exception):
    """An exception for handling F prime settings misconfiguration"""



"""
@file time_tag.py
@brief Class used to parse and store time tags sent with serialized data

This Class is used to parse, store, and create human readable strings for the
time tags sent with serialized data in the fprime architecture.

@date Created Dec 16, 2015
@author: dinkel

@date Updated June 18, 2018
@author R. Joseph Paetz (rpaetz@jpl.nasa.gov)

@date Updated July 22, 2019
@author Kevin C Oran (kevin.c.oran@jpl.nasa.gov)

@bug No known bugs
"""

import datetime
import math
from enum import Enum

# Custom Python Modules
import fprime.common.models.serialize.numerical_types
from fprime.common.models.serialize import type_base
from fprime.common.models.serialize.type_exceptions import TypeRangeException

TimeBase = Enum(
    "TimeBase",
    # No time base has been established
    {
        "TB_NONE": 0,
        # Processor cycle time. Not tried to external time
        "TB_PROC_TIME": 1,
        # Time on workstation where software is running. For testing.
        "TB_WORKSTATION_TIME": 2,
        # Time as reported by the SCLK.
        "TB_SC_TIME": 3,
        # Time as reported by the FPGA clock
        "TB_FPGA_TIME": 4,
        # Don't care value for sequences
        "TB_DONT_CARE": 0xFFFF,
    },
)


class TimeType(type_base.BaseType):
    """
    Representation of the time type

    Used to parse, store, and create human readable versions of the time tags
    included in serialized output from fprime_gds systems

    Note: comparisons support comparing to numbers or other instances of TimeType. If comparing to
    another TimeType, these comparisons use the provided compare method. See TimeType.compare for
    a description of this behavior.  See comparison functions at the end.
    """

    def __init__(self, time_base=0, time_context=0, seconds=0, useconds=0):
        """
        Constructor

        Args
            time_base (int): Time base index for the time tag. Must be a valid
                             integer for a TimeBase Enum value.
            time_context (int): Time context for the time tag
            seconds (int): Seconds elapsed since specified time base
            useconds (int): Microseconds since start of current second. Must
                            be in range [0, 999999] inclusive

        Returns:
            An initialized TimeType object
        """
        # Layout of time tag:
        #
        # START (LSB)
        # |  2 bytes  |    1 byte    | 4 bytes |   4 bytes    |
        # |-----------|--------------|---------|--------------|
        # | Time Base | Time Context | Seconds | Microseconds |
        super().__init__()
        self._check_time_base(time_base)
        self._check_useconds(useconds)

        self.__timeBase = fprime.common.models.serialize.numerical_types.U16Type(
            time_base
        )
        self.__timeContext = fprime.common.models.serialize.numerical_types.U8Type(
            time_context
        )
        self.__secs = fprime.common.models.serialize.numerical_types.U32Type(seconds)
        self.__usecs = fprime.common.models.serialize.numerical_types.U32Type(useconds)

    @staticmethod
    def _check_useconds(useconds):
        """
        Checks if a given microsecond value is valid.

        Args:
            usecs (int): The value to check

        Returns:
            None if valid, raises TypeRangeException if not valid.
        """
        if (useconds < 0) or (useconds > 999999):
            raise TypeRangeException(useconds)

    @staticmethod
    def _check_time_base(time_base):
        """
        Checks if a given TimeBase value is valid.

        Args:
            time_base (int): The value to check

        Returns:
            Returns if valid, raises TypeRangeException if not valid.
        """
        # Construct list of possible values for TimeBase enum
        valid_vals = [member.value for member in list(TimeBase)]

        if time_base not in valid_vals:
            raise TypeRangeException(time_base)

    def to_jsonable(self):
        """
        JSONable object format
        """
        return {
            "type": self.__repr__(),
            "base": self.__timeBase,
            "context": self.__timeContext,
            "seconds": self.seconds,
            "microseconds": self.useconds,
        }

    @property
    def timeBase(self):
        return TimeBase(self.__timeBase.val)

    @timeBase.setter
    def timeBase(self, val):
        self._check_time_base(val)
        self.__timeBase = fprime.common.models.serialize.numerical_types.U16Type(val)

    @property
    def timeContext(self):
        return self.__timeContext.val

    @timeContext.setter
    def timeContext(self, val):
        self.__timeContext = fprime.common.models.serialize.numerical_types.U8Type(val)

    @property
    def seconds(self):
        return self.__secs.val

    @seconds.setter
    def seconds(self, val):
        self.__secs = fprime.common.models.serialize.numerical_types.U32Type(val)

    @property
    def useconds(self):
        return self.__usecs.val

    @useconds.setter
    def useconds(self, val):
        self._check_useconds(val)
        self.__usecs = fprime.common.models.serialize.numerical_types.U32Type(val)

    def serialize(self):
        """
        Serializes the time type

        Returns:
            Byte array containing serialized time type
        """
        buf = b""
        buf += self.__timeBase.serialize()
        buf += self.__timeContext.serialize()
        buf += self.__secs.serialize()
        buf += self.__usecs.serialize()
        return buf

    def deserialize(self, data, offset):
        """
        Deserializes a serialized time type in data starting at offset.

        Internal values to the object are updated.

        Args:
            data: binary data containing the time tag (type = bytearray)
            offset: Index in data where time tag starts
        """

        # Decode Time Base
        self.__timeBase.deserialize(data, offset)
        offset += self.__timeBase.getSize()

        # Decode Time Context
        self.__timeContext.deserialize(data, offset)
        offset += self.__timeContext.getSize()

        # Decode Seconds
        self.__secs.deserialize(data, offset)
        offset += self.__secs.getSize()

        # Decode Microseconds
        self.__usecs.deserialize(data, offset)
        offset += self.__usecs.getSize()

    def getSize(self):
        """
        Return the size of the time type object when serialized

        Returns:
            The size of the time type object when serialized
        """
        return (
            self.__timeBase.getSize()
            + self.__timeContext.getSize()
            + self.__secs.getSize()
            + self.__usecs.getSize()
        )

    @staticmethod
    def compare(t1, t2):
        """
        Compares two TimeType objects

        This function sorts times in the order of: timeBase, secs, usecs, and
        then timeContext.

        Returns:
            Negative, 0, or positive for t1<t2, t1==t2, t1>t2 respectively
        """

        def cmp(x, y):
            return (x > y) - (x < y)  # added to support Python 2/3

        # Compare Base
        base_cmp = cmp(t1.timeBase.value, t2.timeBase.value)
        if base_cmp != 0:
            return base_cmp

        # Compare seconds
        sec_cmp = cmp(t1.seconds, t2.seconds)
        if sec_cmp != 0:
            return sec_cmp

        # Compare usecs
        usec_cmp = cmp(t1.useconds, t2.useconds)
        if usec_cmp != 0:
            return usec_cmp

        # Compare contexts
        return cmp(t1.timeContext, t2.timeContext)

    def __str__(self):
        """
        Formats the time type object for printing

        Returns:
            A string representing the time type object
        """
        return "(%d(%d)-%d:%d)" % (
            self.__timeBase.val,
            self.__timeContext.val,
            self.__secs.val,
            self.__usecs.val,
        )

    def to_readable(self, time_zone=None):
        """
        Returns a string of the time object in a human readable format

        Args:
            time_zone (tzinfo): Time zone to convert the TimeType
                      object to before printing. Timezone also displayed.
                      If time_zone=None, local timezone is used.
                      Defaults to None.

        Returns:
            A human readable string representing the time type object
        """
        dt = self.get_datetime(time_zone)

        # If we could convert to a valid datetime, use that, otherwise, format
        if dt:
            return dt.strftime("%Y-%m-%d %H:%M:%S%z")
        else:
            return "%s: %d.%06ds, context=%d" % (
                TimeBase(self.__timeBase.val).name,
                self.__secs.val,
                self.__usecs.val,
                self.__timeContext.val,
            )

    def get_datetime(self, tz=None):
        """
        Returns the python datetime object for UTC time

        Args:
            tz (tzinfo): timezone to create the datetime object
               in. If tz=None, local time zone used. Defaults to None.
        Returns:
            datetime object for the time type or None if the time couldn't
            be determined.
        """

        tb = TimeBase(self.__timeBase.val)
        dt = None

        if tb in [TimeBase["TB_WORKSTATION_TIME"], TimeBase["TB_SC_TIME"]]:

            # This finds the local time corresponding to the timestamp and
            # timezone object, or local time zone if tz=None
            dt = datetime.datetime.fromtimestamp(self.__secs.val, tz)

            dt = dt.replace(microsecond=self.__usecs.val)

        return dt

    def set_datetime(self, dt, time_base=0xFFFF):
        """
        Sets the timebase from a datetime object.

        Args:
            dt (datetime): datetime object to read from time.
        """
        total_seconds = (dt - datetime.datetime.fromtimestamp(0)).total_seconds()
        seconds = int(total_seconds)
        useconds = int((total_seconds - seconds) * 1000000)

        self._check_time_base(time_base)
        self._check_useconds(useconds)

        self.__timeBase = fprime.common.models.serialize.numerical_types.U16Type(
            time_base
        )
        self.__secs = fprime.common.models.serialize.numerical_types.U32Type(seconds)
        self.__usecs = fprime.common.models.serialize.numerical_types.U32Type(useconds)

    # The following Python special methods add support for rich comparison of TimeTypes to other
    # TimeTypes and numbers.

    def get_float(self):
        """
        a helper method that gets the current TimeType as a float where the non-fraction is seconds
        and the fraction is microseconds. This enables comparisons with numbers.
        """
        return self.seconds + (self.useconds / 1000000)

    def __lt__(self, other):
        """Less than"""
        if isinstance(other, TimeType):
            return self.compare(self, other) < 0
        else:
            return self.get_float() < other

    def __le__(self, other):
        """Less than or equal"""
        if isinstance(other, TimeType):
            return self.compare(self, other) <= 0
        else:
            return self.get_float() <= other

    def __eq__(self, other):
        """Equal"""
        if isinstance(other, TimeType):
            return self.compare(self, other) == 0
        else:
            return self.get_float() == other

    def __ne__(self, other):
        """Not equal"""
        if isinstance(other, TimeType):
            return self.compare(self, other) != 0
        else:
            return self.get_float() != other

    def __gt__(self, other):
        """Greater than"""
        if isinstance(other, TimeType):
            return self.compare(self, other) > 0
        else:
            return self.get_float() > other

    def __ge__(self, other):
        """Greater than or equal"""
        if isinstance(other, TimeType):
            return self.compare(self, other) >= 0
        else:
            return self.get_float() >= other

    # The following helper methods enable support for arithmetic operations on TimeTypes.

    def set_float(self, num):
        """
        a helper method that takes a float and sets a TimeType's seconds and useconds fields.
        Note: This method is private because it is only used by the _get_type_from_float helper to
        generate new TimeType instances. It is not meant to be used to modify an existing timestamp.
        Note: Present implementation will set any negative result to 0
        """
        num = max(num, 0)
        self.seconds = int(math.floor(num))
        self.useconds = int(round((num - self.seconds) * 1000000))

    def get_type_from_float(self, num):
        """
        a helper method that returns a new instance of TimeType and sets the seconds and useconds
        fields using the given number. The new TimeType's time_base and time_context will be
        preserved from the calling object.
        """
        tType = TimeType(self.__timeBase.val, self.__timeContext.val)
        tType.set_float(num)
        return tType

    # The following Python special methods add support for arithmetic operations on TimeTypes.

    def __add__(self, other):
        """Addition"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = self.get_float() + other
        return self.get_type_from_float(num)

    def __sub__(self, other):
        """Subtraction"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = self.get_float() - other
        return self.get_type_from_float(num)

    def __mul__(self, other):
        """Multiplication"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = self.get_float() * other
        return self.get_type_from_float(num)

    def __truediv__(self, other):
        """True division"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = self.get_float() / other
        return self.get_type_from_float(num)

    def __floordiv__(self, other):
        """Floored division"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = self.get_float() // other
        return self.get_type_from_float(num)

    # The following Python special methods add support for reflected arithmetic operations on TimeTypes.

    def __radd__(self, other):
        """Reflected addition"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = other + self.get_float()
        return self.get_type_from_float(num)

    def __rsub__(self, other):
        """Reflected subtraction"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = other - self.get_float()
        return self.get_type_from_float(num)

    def __rmul__(self, other):
        """Reflected multiplication"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = other * self.get_float()
        return self.get_type_from_float(num)

    def __rtruediv__(self, other):
        """Reflected division"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = other / self.get_float()
        return self.get_type_from_float(num)

    def __rfloordiv__(self, other):
        """Reflected floored division"""
        if isinstance(other, TimeType):
            other = other.get_float()
        num = other // self.get_float()
        return self.get_type_from_float(num)



""" fprime.fbuild.cli: fbuild command line processing

Command line processing functions for fbuild (fprime build system targets) of "generate", "purge", and build system
target operations.

@author mstarch
"""
import argparse
from pathlib import Path
from typing import Dict, List, Tuple, Callable
from fprime.fbuild.types import BuildType
from fprime.fbuild.target import Target
from fprime.fbuild.builder import Build
from fprime.fbuild.interaction import confirm


def get_target(parsed: argparse.Namespace) -> Target:
    """Gets the target given the argparse namespace

    Takes the parsed namespace and processes it to a known matching target.

    Args:
        parsed: argparse namespace to read values from

    Returns:
        target that can support the given supplied namespace
    """
    mnemonic = parsed.command
    flags = {
        flag for flag in Target.get_all_possible_flags() if getattr(parsed, flag, False)
    }
    return Target.get_target(mnemonic, flags)


def run_fbuild_cli(
    build: Build,
    parsed: argparse.Namespace,
    cmake_args: Dict[str, str],
    make_args: Dict[str, str],
    pass_through: List[str],
):
    """Execution of the fbuild commands

    Args:
        build: build object, pre-loaded, iff not "generate" or "purged"
        parsed: parsed arguments object.
        cmake_args: cmake commands from parsed output
        make_args: arguments to the make system
    """
    if parsed.command == "generate":
        toolchain = build.find_toolchain()
        print(f"[INFO] Generating build directory at: {build.build_dir}")
        print(f"[INFO] Using toolchain file {toolchain} for platform {parsed.platform}")
        if toolchain is not None:
            cmake_args["CMAKE_TOOLCHAIN_FILE"] = toolchain
        build.generate(cmake_args)
    elif parsed.command == "purge":
        # Since purge does not load its "base", we need to overload the platform
        build.platform = parsed.platform
        for purge_build in Build.get_build_list(
            build, parsed.build_cache, ignore_invalid=parsed.force
        ):
            print(
                f"[INFO] {parsed.command.title()} build directory at: {purge_build.build_dir}"
            )
            if parsed.force or confirm("Purge this directory (yes/no)?"):
                purge_build.purge()
            install_dir = purge_build.install_dest_exists()
            if (
                purge_build.build_type != BuildType.BUILD_CUSTOM
                and install_dir
                and install_dir.exists()
            ):
                print(
                    f"[INFO] {parsed.command.title()} install directory at: {install_dir}"
                )
                if parsed.force or confirm(
                    f"Purge installation directory at {install_dir} (yes/no)?"
                ):
                    purge_build.purge_install()
    else:
        target = get_target(parsed)
        target.execute(build, context=Path(parsed.path), args=(make_args, pass_through))


def add_target_parser(
    target: Target,
    subparsers,
    common: argparse.ArgumentParser,
    existing: Dict[str, Tuple[argparse.ArgumentParser, List[str]]],
    help_text: "HelpText",
) -> str:
    """Add a subparser for a given build target

    For a given build target, construct an argument parser with mnemonic as subcommand and needed flags. Then add it as
    a subparser to the global parent. If it already exists in existing, then just add non-existent flags/

    Args:
        target:     target for building a new subparser
        subparsers: argument parser to add a subparser to
        common:     common subparser to be used as parent carrying common flags
        existing:   dictionary storing the mnemonic to parser and flags tuple
        help_text:  mnemonic to help text mapping

    Returns:
        Name of the command this parser handles

    Notes:
        This functions has side effects of editing existing and the list of subparsers
    """
    help_string = help_text.short(
        target.mnemonic, f"{target.desc} in the specified directory"
    )
    description = help_text.long(target.mnemonic, help_string)
    if target.mnemonic not in existing:
        parser = subparsers.add_parser(
            target.mnemonic,
            parents=[common],
            add_help=False,
            description=description,
            help=help_string,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        # --ut flag also exists at the global parsers, skip adding it
        existing[target.mnemonic] = (parser, ["ut"])
        # Common target-only items
        parser.add_argument(
            "-j",
            "--jobs",
            default=1,
            type=int,
            help="Parallel build job count. Default: %(default)s.",
        )
    parser, flags = existing[target.mnemonic]
    new_flags = [flag for flag in target.flags if flag not in flags]
    for flag in new_flags:
        extra_help = (
            f". Use '--pass-through' to supply '{target.pass_handler()}' arguments"
            if target.allows_pass_args()
            else ""
        )
        parser.add_argument(
            f"--{flag}",
            action="store_true",
            default=False,
            help=f"{target.desc}{extra_help}",
        )
    # Allow pass through arguments
    if target.pass_handler() and "--pass-through" not in flags:
        parser.add_argument(
            "--pass-through",
            nargs=argparse.REMAINDER,
            default=[],
            help="If specified, --pass-through must be the last argument. Remaining arguments passed to underlying executable",
        )
        flags.append("--pass-through")
    flags.extend(new_flags)
    return target.mnemonic


def add_special_targets(
    subparsers, common: argparse.ArgumentParser, help_text: "HelpText"
) -> List[str]:
    """Add in generate and purge commands

    Args:
        subparsers: subparsers used to create new subparsers
        common: common parser
        help_text: map of mnemonic to help text

    Returns:
        ["generate", "purge"] list of special targets
    """

    generate_parser = subparsers.add_parser(
        "generate",
        help=help_text.short(
            "generate", "Generate a build cache for specified deployment"
        ),
        description=help_text.long(
            "generate", "Generate a build cache for specified deployment"
        ),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    generate_parser.add_argument(
        "-Dxyz",
        action="append",
        help="Pass -D flags through to CMakes",
        nargs=1,
        default=[],
    )
    purge_parser = subparsers.add_parser(
        "purge",
        help=help_text.short("purge", "Remove build caches for specified deployment"),
        description=help_text.long(
            "purge", "Remove build caches for specified deployment"
        ),
        add_help=False,
        parents=[common],
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    purge_parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Purges the build directory by force. No confirmation will be requested.",
    )
    return ["generate", "purge"]


def add_fbuild_parsers(
    subparsers, common: argparse.ArgumentParser, help_text: "HelpText"
) -> Tuple[Dict[str, Callable], Dict[str, argparse.ArgumentParser]]:
    """Add in the build-system targets: generate, purge, and all build endpoints

    Sets up the targets used to generate and run the build system. This includes the purge command for cleaning up. This
    wraps up all targets defined in the targets list in fprime.fbuild.builder

    Args:
        subparsers: subparsers object used to add subparsers to the global CLI
        common: common parser used to add all common arguments
        help_text: map of mnemonics to help text

    Returns:
        Tuple of two dictionaries mapping command name to the function used to process commands and to parser
    """
    parsers = {}
    run_map = {
        add_target_parser(
            target, subparsers, common, parsers, help_text=help_text
        ): run_fbuild_cli
        for target in Target.get_all_targets()
    }
    run_map.update(
        {
            command: run_fbuild_cli
            for command in add_special_targets(subparsers, common, help_text=help_text)
        }
    )
    return run_map, parsers



"""
Created on Dec 18, 2014

@author: reder
Replaced type base class with decorators
"""
import abc

from .type_exceptions import AbstractMethodException


class BaseType(abc.ABC):
    """
    An abstract base defining the methods supported by all base classes.
    """

    @abc.abstractmethod
    def serialize(self):
        """
        Serializes the current object type.
        """
        raise AbstractMethodException("serialize")

    @abc.abstractmethod
    def deserialize(self, data, offset):
        """
        AbstractDeserialize interface
        """
        raise AbstractMethodException("deserialize")

    @abc.abstractmethod
    def getSize(self):
        """
        Abstract getSize interface
        """
        raise AbstractMethodException("getSize")

    def __repr__(self):
        """Produces a string representation of a given type"""
        return self.__class__.__name__.replace("Type", "")

    @abc.abstractmethod
    def to_jsonable(self):
        """
        Converts this type to a JSON serializable object
        """
        raise AbstractMethodException("to_jsonable")


class ValueType(BaseType):
    """
    An abstract base type used to represent a single value. This defines the value property, allowing for setting and
    reading from the .val member.
    """

    def __init__(self, val=None):
        """Defines the single value"""
        self._val = None
        # Run full setter
        if val is not None:
            self.val = val

    @classmethod
    @abc.abstractmethod
    def validate(cls, val):
        """
        Checks the val for validity with respect to the current type. This will raise TypeMismatchException when the
        validation fails of the val's type fails. It will raise TypeRangeException when val is out of range. Concrete
        implementations will raise other exceptions based on type. For example, serializable types raise exceptions for
        missing members.

        :param val: value to validate
        :raises TypeMismatchException: value has incorrect type, TypeRangeException: val is out of range
        """

    def __eq__(self, other):
        """Check equality between types"""
        return False if type(other) != type(self) else self._val == other._val

    @property
    def val(self):
        """Getter for .val"""
        return self._val

    @val.setter
    def val(self, val):
        """Setter for .val calls validate internally"""
        self.validate(val)
        self._val = val

    def to_jsonable(self):
        """
        Converts this type to a JSON serializable object
        """
        return {"value": self.val, "type": str(self.__class__)}


class DictionaryType(ValueType, abc.ABC):
    """Type whose specification is defined in the dictionary

    Certain types in fprime (strings, serializables, enums) are defined in the dictionary. Where all projects have
    access to primitive types (U8, F32, etc) and the definitions of theses types is global, other types complete
    specification comes from the dictionary itself. String set max-lengths per project, serializable fields are defined,
    and enumeration values are enumerated. This class is designed to take base complex types (StringType, etc) and build
    dynamic subclasses for the given dictionary defined type.
    """

    _CONSTRUCTS = {}

    @classmethod
    def construct_type(cls, parent_class, name, **class_properties):
        """Construct a new dictionary type

        Construct a new dynamic subtype of the given base type. This type will be named with the name parameter, define
        the supplied class properties, and will be a subtype of the parent class. This function registers these new
        types by name in the DictionaryType._CONSTRUCTS dictionary. When a type is defined a second or more times, the
        originally constructed sub type is used and the newly supplied class properties are validated for equality
        against the original definition. It is an error to define the type multiple times with inconsistent class
        properties.

        Args:
            cls: DictionaryType, used to access the class property _CONSTRUCTS
            parent_class: class used as parent to the new sub type
            name: name of the new sub type
            **class_properties: properties to define on the subtype (e.g. max length for strings)
        """
        assert (
            parent_class != DictionaryType
        ), "Cannot build dictionary type from dictionary type directly"
        construct, original_properties = cls._CONSTRUCTS.get(
            name, (type(name, (parent_class,), class_properties), class_properties)
        )
        # Validate both new properties against original properties and against what is set on original class
        assert (
            original_properties == class_properties
        ), "Different class properties specified"
        for attribute, value in class_properties.items():
            previous_value = getattr(construct, attribute, None)
            assert (
                previous_value == value
            ), f"Class {name} differs by attribute {attribute}. {previous_value} vs {value}"
        cls._CONSTRUCTS[name] = (construct, class_properties)
        return construct



"""
Created on Dec 18, 2014
@author: tcanham, reder
"""
import struct

from .type_base import DictionaryType
from .type_exceptions import (
    DeserializeException,
    EnumMismatchException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
)


class EnumType(DictionaryType):
    """
    Representation of the ENUM type.

    The enumeration takes a dictionary that stores the enumeration members
    and current value as a string. The member values will have to be computed
    containing code based on C enum rules
    """

    @classmethod
    def construct_type(cls, name, enum_dict):
        """Construct the custom enum type

        Constructs the custom enumeration type, with the supplied enumeration dictionary.

        Args:
            name: name of the enumeration type
            enum_dict: enumeration: value dictionary defining the enumeration
        """
        if not isinstance(enum_dict, dict):
            raise TypeMismatchException(dict, type(enum_dict))
        for member in enum_dict.keys():
            if not isinstance(member, str):
                raise TypeMismatchException(str, type(member))
            if not isinstance(enum_dict[member], int):
                raise TypeMismatchException(int, enum_dict[member])
        return DictionaryType.construct_type(cls, name, ENUM_DICT=enum_dict)

    @classmethod
    def validate(cls, val):
        """Validate the value passed into the enumeration"""
        if not isinstance(val, str):
            raise TypeMismatchException(str, type(val))
        if val not in cls.keys():
            raise EnumMismatchException(cls.__class__.__name__, val)

    @classmethod
    def keys(cls):
        """
        Return all the enum key values.
        """
        return list(cls.ENUM_DICT.keys())

    def serialize(self):
        """
        Serialize the enumeration type using an int type
        """
        # for enums, take the string value and convert it to
        # the numeric equivalent
        if self._val is None or (
            self._val == "UNDEFINED" and "UNDEFINED" not in self.ENUM_DICT
        ):
            raise NotInitializedException(type(self))
        return struct.pack(">i", self.ENUM_DICT[self._val])

    def deserialize(self, data, offset):
        """
        Deserialize the enumeration using an int type
        """
        try:
            int_val = struct.unpack_from(">i", data, offset)[0]
        except struct.error:
            raise DeserializeException(
                f"Could not deserialize enum value. Needed: {self.getSize()} bytes Found: {len(data[offset:])}"
            )
        for key, val in self.ENUM_DICT.items():
            if int_val == val:
                self._val = key
                break
        # Value not found, invalid enumeration value
        else:
            raise TypeRangeException(int_val)

    def getSize(self):
        """Calculates the size based on the size of an integer used to store it"""
        return struct.calcsize(">i")



""" Cookie cutter wrapper used to template out components
"""
import os
import glob
import sys
import textwrap
from pathlib import Path
from typing import Dict
import re
from contextlib import contextmanager

from cookiecutter.main import cookiecutter
from cookiecutter.exceptions import OutputDirExistsException
from jinja2 import Environment, FileSystemLoader

from fprime.fbuild.builder import Build
from fprime.fbuild.target import Target
from fprime.fbuild.cmake import CMakeExecutionException


def confirm(msg):
    """Confirms the removal of the file with a yes or no input"""
    # Loop "forever" intended
    while True:
        confirm_input = input(msg)
        if confirm_input.lower() in ["y", "yes"]:
            return True
        if confirm_input.lower() in ["n", "no"]:
            return False
        print(f"{confirm_input} is invalid.  Please use 'yes' or 'no'")


def replace_contents(filename, what, replacement, count=1):
    changelog = Path(filename).read_text()
    with open(filename, "w") as fh:
        new_file = changelog.replace(what, replacement, count)
        fh.write(new_file)
        return new_file != changelog


def run_impl(deployment: Path, path: Path, platform: str, verbose: bool):
    """Run implementation of files one time"""
    target = Target.get_target("impl", set())
    build = Build(target.build_type, deployment, verbose=verbose)
    build.load(platform)

    hpp_files = glob.glob(f"{path}/*.hpp", recursive=False)
    cpp_files = glob.glob(f"{path}/*.cpp", recursive=False)
    cpp_files.sort(key=len)

    # Check destinations
    if not hpp_files or not cpp_files:
        print(
            "[WARNING] Failed to find .cpp and .hpp destination files for implementation."
        )
        return False

    common = [name for name in cpp_files if "Common" in name] + []
    hpp_dest = hpp_files[0]
    cpp_dest = common[0] if common else cpp_files[0]

    if not confirm(
        f"Generate implementations and merge into {hpp_dest} and {cpp_dest}?"
    ):
        return False
    print("Generating implementation files and merging...")
    with suppress_stdout():
        build.execute(target, context=path, make_args={})

    hpp_files_template = glob.glob(f"{path}/*.hpp-template", recursive=False)
    cpp_files_template = glob.glob(f"{path}/*.cpp-template", recursive=False)

    if not hpp_files_template or not cpp_files_template:
        print("[WARNING] Failed to find generated .cpp-template or .hpp-template files")
        return False

    hpp_src = hpp_files_template[0]
    cpp_src = cpp_files_template[0]

    # Copy files without headers
    for src, dest in [(hpp_src, hpp_dest), (cpp_src, cpp_dest)]:
        with open(src, "r") as file_handle:
            lines = file_handle.readlines()
        lines = lines[11:]  # Remove old header
        with open(dest, "a") as file_handle:
            file_handle.write("".join(lines))

    removals = [Path(hpp_src), Path(cpp_src)]
    for removal in removals:
        os.remove(removal)
    return True


def add_to_cmake(list_file: Path, comp_path: Path):
    """Adds new component or port to CMakeLists.txt"""
    print(f"[INFO] Found CMakeLists.txt at '{list_file}'")
    with open(list_file, "r") as f:
        lines = f.readlines()

    addition = (
        'add_fprime_subdirectory("${CMAKE_CURRENT_LIST_DIR}/' + str(comp_path) + '/")\n'
    )
    if addition in lines:
        print("Already added to CMakeLists.txt")
        return True

    if not confirm(f"Add component {comp_path} to {list_file} at end of file?"):
        return False

    lines.append(addition)
    with open(list_file, "w") as f:
        f.write("".join(lines))
    return True


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def regenerate(build: Build):
    print("Refreshing cache to include new addition...")
    with suppress_stdout():
        try:
            build.cmake.cmake_refresh_cache(build.get_build_cache())
        except CMakeExecutionException:
            build.cmake.cmake_refresh_cache(build.get_build_cache(), True)


def add_unit_tests(deployment, comp_path, platform, verbose):
    # Creates unit tests and moves them into test/ut directory
    os.chdir(str(comp_path))
    if confirm("Would you like to generate unit tests?: "):
        test_path = Path("test", "ut")
        test_path.mkdir(parents=True, exist_ok=True)
        target = Target.get_target("impl", {"ut"})
        build = Build(target.build_type, deployment, verbose=verbose)
        build.load(platform)
        print("Generating unit tests...")
        with suppress_stdout():
            build.execute(target, context=comp_path, make_args={})
        test_files = [
            "Tester.hpp",
            "Tester.cpp",
            "TesterBase.hpp",
            "TesterBase.cpp",
            "GTestBase.hpp",
            "GTestBase.cpp",
            "TestMain.cpp",
        ]
        for file in test_files:
            if os.path.isfile(file):
                new_name = test_path / file
                os.rename(file, str(new_name))

        with open("CMakeLists.txt", "r") as f:
            cmakeFile = f.read()

        with open("CMakeLists.txt", "w") as f:
            cmakeFile = cmakeFile + textwrap.dedent(
                """\n
                set(UT_SOURCE_FILES
                  "${CMAKE_CURRENT_LIST_DIR}/test/ut/TestMain.cpp"
                  "${CMAKE_CURRENT_LIST_DIR}/test/ut/Tester.cpp"
                )

                register_fprime_ut()"""
            )
            f.write(cmakeFile)

        if replace_contents(
            Path("test", "ut", "Tester.hpp"), "ComponentImpl.hpp", ".hpp", -1
        ):
            print("[INFO] Fixed hpp include in Tester.hpp")
        print("[INFO] Unit tests were generated.")


def add_port_to_cmake(list_file: Path, comp_path: Path):
    """Adds new port to CMakeLists.txt in port directory"""
    print(f"[INFO] Found CMakeLists.txt at '{list_file}'")
    with open(list_file, "r") as file_handle:
        lines = file_handle.readlines()
    index = 0
    while re.search("set\(\s*SOURCE_FILES", lines[index]) is None:
        index += 1
    index += 1
    while "CMAKE_CURRENT_LIST_DIR" in lines[index]:
        index += 1
    if not confirm(f"Add port {comp_path} to {list_file} ports in CMakeLists.txt?"):
        return False

    addition = '    "${{CMAKE_CURRENT_LIST_DIR}}/{}"\n'.format(comp_path)
    lines.insert(index, addition)
    with open(list_file, "w") as file_handle:
        file_handle.write("".join(lines))
    return True


def find_nearest_cmake_lists(component_dir: Path, deployment: Path, proj_root: Path):
    """Find the nearest CMakeLists.txt file

    The "nearest" file is defined as the closes parent that is not the "project root" that contains a CMakeLists.txt. If
    none is found, the same procedure is run from the deployment directory and includes the project root this time. If
    nothing is found, None is returned.

    In short the following in order of preference:
     - Any Component Parent
     - Any Deployment Parent
     - Project Root
     - None

    Args:
        component_dir: directory of new component
        deployment: deployment directory
        proj_root: project root directory

    Returns:
        path to CMakeLists.txt or None
    """
    test_path = component_dir.parent
    # First iterate from where we are, then from the deployment to find the nearest CMakeList.txt nearby
    for test_path, end_path in [(test_path, proj_root), (deployment, proj_root.parent)]:
        while proj_root is not None and test_path != proj_root.parent:
            cmake_list_file = test_path / "CMakeLists.txt"
            if cmake_list_file.is_file():
                return cmake_list_file
            test_path = test_path.parent
    return None


def new_component(deployment: Path, platform: str, verbose: bool, build: Build):
    """Uses cookiecutter for making new components"""
    try:
        print("[WARNING] **** fprime-util new is prototype functionality ****")
        proj_root = build.get_settings("project_root", None)

        # Checks if component_cookiecutter is set in settings.ini file, else uses local component_cookiecutter template as default
        if (
            build.get_settings("component_cookiecutter", None) is not None
            and build.get_settings("component_cookiecutter", None) != "default"
        ):
            source = build.get_settings("component_cookiecutter", None)
            print(f"[INFO] Cookiecutter source: {source}")
        else:
            source = (
                os.path.dirname(__file__)
                + "/../cookiecutter_templates/cookiecutter-fprime-component"
            )
            print("[INFO] Cookiecutter source: using builtin")
        print()
        final_component_dir = Path(
            cookiecutter(source, extra_context={"component_namespace": deployment.name})
        ).resolve()
        if proj_root is None:
            print(
                f"[INFO] Created component directory without adding to build system nor generating implementation {final_component_dir}"
            )
            return 0
        # Attempt to register to CMakeLists.txt
        proj_root = Path(proj_root)
        cmake_lists_file = find_nearest_cmake_lists(
            final_component_dir, deployment, proj_root
        )
        if cmake_lists_file is None or not add_to_cmake(
            cmake_lists_file, final_component_dir.relative_to(cmake_lists_file.parent)
        ):
            print(
                f"[INFO] Could not register {final_component_dir} with build system. Please add it and generate implementations manually."
            )
            return 0
        regenerate(build)
        # Attempt implementation
        if not run_impl(deployment, final_component_dir, platform, verbose):
            print(
                "[INFO] Could not generate implementations. Please do so manually.".format(
                    final_component_dir
                )
            )
            return 0
        cpp_file = glob.glob(str(Path(deployment.name, final_component_dir, "*.cpp")))[
            0
        ]
        print("[INFO] Created new component and created initial implementations.")
        if replace_contents(cpp_file, "ComponentImpl.hpp", ".hpp", -1):
            print("[INFO] Fixed hpp include in cpp file.")

        add_unit_tests(deployment, final_component_dir, platform, verbose)
        print(
            f'[INFO] Next run `fprime-util build{"" if platform is None else f" {platform}"}` in {final_component_dir}'
        )
        return 0
    except OutputDirExistsException as out_directory_error:
        print(f"{out_directory_error}", file=sys.stderr)
    except CMakeExecutionException as exc:
        print(f"[ERROR] Failed to create component. {exc}", file=sys.stderr)
    except OSError as ose:
        print(f"[ERROR] {ose}")
    return 1


def is_valid_name(word: str):
    invalid_characters = [
        "#",
        "%",
        "&",
        "{",
        "}",
        "/",
        "\\",
        "<",
        ">",
        "*",
        "?",
        " ",
        "$",
        "!",
        "'",
        '"',
        ":",
        "@",
        "+",
        "`",
        "|",
        "=",
    ]
    for char in invalid_characters:
        if isinstance(word, str) and char in word:
            return char
        elif not isinstance(word, str):
            raise ValueError("Incorrect usage of is_valid_name")
    return "valid"


def get_valid_input(prompt):
    valid_name = False
    while not valid_name:
        name = input(prompt)
        char = is_valid_name(name)
        if char != "valid":
            print("'" + char + "' is not a valid character")
        else:
            valid_name = True
    return name


def get_port_input(namespace):
    # Gather inputs to use as context for the port template
    defaults = {
        "port_name": "ExamplePort",
        "short_description": "Example usage of port",
        "dir_name": ".",
        "namespace": namespace,
        "arg_list": [],
    }
    args_done = False
    arg_list = []
    port_name = get_valid_input(f'Port Name [{defaults["port_name"]}]: ')
    short_description = input(f'Short Description [{defaults["short_description"]}]: ')
    dir_name = get_valid_input(f'Directory Name [{defaults["dir_name"]}]: ')
    namespace = get_valid_input(f'Port Namespace [{defaults["namespace"]}]: ')
    while not args_done:
        add_arg = (
            confirm("Would you like to add another argument?: ")
            if arg_list
            else confirm("Would you like to add an argument?: ")
        )
        if add_arg:
            arg_name = get_valid_input("Argument name: ")
            arg_type = get_valid_input(
                "Argument type (Valid primitive types are: I8, I16, "
                + "I32, U8, U16, U32, F32, F64, NATIVE_INT_TYPE, NATIVE_UNIT_TYPE, and POINTER_CAST. "
                + "You may also use your own user-defined types): "
            )
            arg_description = input("Short description of argument: ")
            arg_list.append((arg_name, arg_type, arg_description))
        else:
            args_done = True
    values = {
        "port_name": port_name,
        "short_description": short_description,
        "dir_name": dir_name,
        "namespace": namespace,
        "arg_list": arg_list,
    }

    # Fill in blank values with defaults
    for key in values:
        if values[key] == "":
            values[key] = defaults[key]
    return values


def new_port(deployment: Path, build: Build):
    """Uses cookiecutter for making new ports"""
    try:
        proj_root = build.get_settings("project_root", None)
        if proj_root is not None:
            proj_root = Path(proj_root)
            proj_root_found = True
        else:
            proj_root_found = False

        PATH = os.path.dirname(os.path.abspath(__file__))
        TEMPLATE_ENVIRONMENT = Environment(
            autoescape=False,
            loader=FileSystemLoader(os.path.join(PATH, "../cookiecutter_templates")),
            trim_blocks=False,
        )
        context = get_port_input(deployment.name)

        if Path(context["dir_name"]).resolve() == deployment.resolve():
            print("[ERROR] cannot create port in deployment directory")
            return 0
        fname = context["port_name"] + "Port" + "Ai.xml"
        if os.path.isfile(Path(context["dir_name"], fname)):
            print(
                "[ERROR] Port",
                context["port_name"],
                "already exists in directory",
                context["dir_name"],
            )
            return 0
        with open(fname, "w") as f:
            xml_file = TEMPLATE_ENVIRONMENT.get_template("port_template.xml").render(
                context
            )
            f.write(xml_file)
        if not os.path.isdir(context["dir_name"]):
            os.mkdir(context["dir_name"])

        os.rename(fname, str(Path(context["dir_name"], fname)))
        path_to_cmakelists = Path(context["dir_name"], "CMakeLists.txt")
        if not os.path.isfile(str(path_to_cmakelists)):
            with open(str(path_to_cmakelists), "w") as f:
                CMake_file = TEMPLATE_ENVIRONMENT.get_template(
                    "CMakeLists_template.txt"
                ).render(context)
                f.write(CMake_file)
        else:
            add_port_to_cmake(str(path_to_cmakelists), fname)

        if proj_root_found is False:
            print(
                "[INFO] No project root found. Created port without adding to build system nor generating implementation."
            )
            return 0
        cmake_lists_file = find_nearest_cmake_lists(
            Path(context["dir_name"]).resolve(), deployment, proj_root
        )
        if cmake_lists_file is None or not add_to_cmake(
            cmake_lists_file,
            (Path(context["dir_name"]).resolve()).relative_to(cmake_lists_file.parent),
        ):
            print(
                f'[INFO] Could not register {Path(context["dir_name"]).resolve()} with build system. Please add it and generate implementations manually.'
            )
            return 0
        regenerate(build)
        print("")
        print(
            "################################################################################"
        )
        print("")
        print(
            "You have successfully created the port "
            + context["port_name"]
            + " located in "
            + context["dir_name"]
        )
        print("")
        print(
            "################################################################################"
        )
        return 0
    except OutputDirExistsException as out_directory_error:
        print(f"{out_directory_error}", file=sys.stderr)
    except CMakeExecutionException as exc:
        print(f"[ERROR] Failed to create port. {exc}", file=sys.stderr)
    except OSError as ose:
        print(f"[ERROR] {ose}")
    return 1



"""
fprime.util.build_helper.py

This script is defined to help users run standard make processes using CMake. This will support migrants from the old
make system, as well as enable more efficient practices when developing in the CMake system. The following functions
are supported herein:

 - build: build the current directory/module/deployment
 - impl: make implementation templates
 - testimpl: make testing templates
 - build_ut: build the current UTs
 - check: run modules unit tests

@author mstarch
"""
import argparse
import os
import re
import sys
from pathlib import Path

from fprime.fbuild.types import InvalidBuildCacheException
from fprime.fbuild.target import Target, NoSuchTargetException
from fprime.fbuild.builder import (
    Build,
    BuildType,
    GenerateException,
    UnableToDetectDeploymentException,
)
from .help_text import HelpText
from .versioning import get_version, VersionException
from fprime.fbuild.cli import add_fbuild_parsers, get_target
from fprime.fpp.cli import add_fpp_parsers
from fprime.util.cli import add_special_parsers

CMAKE_REG = re.compile(r"-D([a-zA-Z0-9_]+)=(.*)")


class ArgValidationException(Exception):
    """An exception used for argument validation"""


def package_version_check(package: str, requirement_path: Path):
    """Checks the version of the packages installed match the expected packages of the fprime aggregate package"""
    import pkg_resources

    expected_version = get_version(package, requirement_path).lstrip(
        "v"
    )  # Python version
    try:
        version = pkg_resources.get_distribution(package).version
        if version != expected_version:
            print(
                f"[WARNING] {package} has unexpected version. Expected: {expected_version} found {version}"
            )
    except pkg_resources.DistributionNotFound:
        print(f"[WARNING] {package} is not installed")


def validate_tools_from_requirements(build: Build):
    """Uses build settings to find requirements file and validate the correct versions installed"""
    # Find prioritized order of requirements.txt, ensuring that each member exists
    possibilities = [
        build.settings.get("project_root", None),
        build.settings.get("framework_path", None),
    ]
    possibilities = [
        Path(possible) / "requirements.txt"
        for possible in possibilities
        if possible is not None
    ]
    possibilities = [possible for possible in possibilities if possible.exists()]
    # Skip tools check, as not requirements.txt found
    if not possibilities:
        print(
            f"[WARNING] Could not find 'requirements.txt' in: {possibilities}. Will not check tool versions."
        )
        return
    # Pre-roll import errors
    try:
        import pkg_resources
    except ImportError:
        print("[WARNING] Cannot import 'pkg_resources'. Will not check tool versions.")
        return

    # Now check each required tool for fprime
    tools = ["fprime-fpp", "fprime-tools", "fprime-gds"]
    for tool in tools:
        for possible in possibilities:
            try:
                package_version_check(tool, possible)
                break
            except (OSError, VersionException) as exc:
                message = f"[WARNING] {exc}"
        else:
            print(message)


def validate(parsed, unknown):
    """
    Validate rules to ensure that the args are properly consistent. This will also generate a set of validated arguments
    to pass to CMake. This allows these values to be created, defaulted, and validated in one place
    :param parsed: args to validate
    :param unknown: unknown arguments
    :return: cmake arguments to pass to CMake
    """
    cmake_args = {}
    make_args = {}
    # Check platforms for existing toolchain, unless the default is specified.
    if not hasattr(parsed, "command") or parsed.command is None:
        raise ArgValidationException("'fprime-util' not supplied sub-command argument")
    if parsed.command == "generate":
        d_args = {
            match.group(1): match.group(2)
            for match in [CMAKE_REG.match(arg) for arg in unknown]
        }
        cmake_args.update(d_args)
        unknown = [arg for arg in unknown if not CMAKE_REG.match(arg)]
    # Build type only for generate, jobs only for non-generate
    elif parsed.command in Target.get_all_targets():
        parsed.settings = None  # Force to load from cache if possible
        make_args["--jobs"] = 1 if parsed.jobs <= 0 else parsed.jobs
    # Check if any arguments are still unknown
    if unknown:
        runnable = f"{os.path.basename(sys.argv[0])} {parsed.command}"
        raise ArgValidationException(
            f"'{runnable}' supplied invalid arguments: {','.join(unknown)}"
        )
    parsed.build_cache = (
        None if parsed.build_cache is None else Path(parsed.build_cache)
    )
    return cmake_args, make_args


def parse_args(args):
    """
    Parse the arguments to the CLI. This will then enable the user to run the above listed commands via the commands.
    :param args: CLI arguments to process
    :return: parsed arguments in a Namespace
    """
    # Common parser specifying common arguments input into the utility
    common_parser = argparse.ArgumentParser(
        description="Common Parser for Common Ingredients."
    )
    common_parser.add_argument(
        "platform",
        nargs="?",
        default="default",
        help="F prime build platform (e.g. Linux, Darwin). Default specified in settings.ini",
    )
    common_parser.add_argument(
        "-d",
        "--deploy",
        dest="deploy",
        default=None,
        help="F prime deployment directory to use. May contain multiple build directories.",
    )
    common_parser.add_argument(
        "-p",
        "--path",
        default=os.getcwd(),
        help="F prime directory to operate on. Default: cwd, %(default)s.",
    )
    common_parser.add_argument(
        "--build-cache",
        dest="build_cache",
        default=None,
        help="Overrides the build cache with a specific directory",
    )
    common_parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Turn on verbose output.",
    )
    common_parser.add_argument(
        "--ut", action="store_true", help="Run command against unit testing build type"
    )

    # Main parser for the whole application
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=HelpText.long("global_help"),
    )
    subparsers = parser.add_subparsers(
        description=HelpText.long("subparsers_description"), dest="command"
    )

    # Add all externally defined cli parser command to running functions
    runners = {}
    parsers = {}

    fbuild_runners, fbuild_parsers = add_fbuild_parsers(
        subparsers, common_parser, HelpText
    )
    fpp_runners, fpp_parsers = add_fpp_parsers(subparsers, common_parser)
    parsers.update(fbuild_parsers)
    parsers.update(fpp_parsers)
    runners.update(fbuild_runners)
    runners.update(fpp_runners)
    runners.update(add_special_parsers(subparsers, common_parser, HelpText))

    # Parse and prepare to run
    parsed, unknown = parser.parse_known_args(args)
    try:
        cmake_args, make_args = validate(parsed, unknown)
    except ArgValidationException as exc:
        print(f"[ERROR] {exc}", end="\n\n")
        parsers.get(parsed.command, (parser,))[0].print_usage()
        sys.exit(1)
    return parsed, cmake_args, make_args, parser, runners


def utility_entry(args):
    """Main interface to F prime utility"""
    parsed, cmake_args, make_args, parser, runners = parse_args(args)

    try:
        try:
            target = get_target(parsed)
            build_type = target.build_type
        except NoSuchTargetException:
            build_type = (
                BuildType.BUILD_TESTING if parsed.ut else BuildType.BUILD_NORMAL
            )

        deployment = (
            Path(parsed.deploy)
            if parsed.deploy is not None
            else Build.find_nearest_deployment(Path.cwd())  # Deployments look in CWD
        )
        build = Build(build_type, deployment, verbose=parsed.verbose)

        # All commands need to load the build cache to setup the basic information for the build with the exception of
        # generate, which is run before the creation of the build cache and thus must invent the cache instead. This
        # call will ensure the build is in a ready state before attempting to check tool versions and run the command.
        #
        # Some commands, like purge and info, run on sets of directories and will attempt to load those sets later.
        # However, the base directory must be setup here. Errors in this load are ignored to allow the command to find
        # build caches related to that set.
        try:
            if parsed.command == "generate":
                build.invent(parsed.platform, build_dir=parsed.build_cache)
            else:
                build.load(parsed.platform, parsed.build_cache)
        except InvalidBuildCacheException:
            if parsed.command not in ["purge", "info"]:
                raise
        validate_tools_from_requirements(build)
        runners[parsed.command](
            build, parsed, cmake_args, make_args, getattr(parsed, "pass_through", [])
        )
    except GenerateException as genex:
        print(
            f"[ERROR] {genex}. Partial build cache remains. Run purge to clean-up.",
            file=sys.stderr,
        )
        return genex.exit_code
    except UnableToDetectDeploymentException:
        print(f"[ERROR] Could not detect deployment directory for: {parsed.path}")
        return 1
    except Exception as exc:
        print(f"[ERROR] {exc}", file=sys.stderr)
        return 1
    return 0



""" fprime.fbuild.target: build target support

Contains the supporting definitions for build targets. These targets are used to run various parts of the build and may
contain build system targets (e.g. CMake target invokers), and miscellaneous targets that perform other actions.

@author lestarch
"""
import functools
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Dict, List, Set, Tuple
from .types import BuildType, NoSuchTargetException


class TargetScope(Enum):
    """Scoping for target execution: GLOBAL, LOCAL

    GLOBAL targets trigger top-level (global) build system targets. LOCAL targets trigger per-directory build system
    targets. BOTH represents a target that can operate in both LOCAL and GLOBAL mode one at a time. When registering
    a BOTH targets, the system will create a local target and a global target and register those. These targets differ
    in both scope and flags as the GLOBAL target receives the flag "--all" added to its list.
    """

    GLOBAL = 0x1
    LOCAL = 0x2
    BOTH = GLOBAL | LOCAL


class ExecutableAction(ABC):
    """Executable action not declaring a formal mnemonic, description, etc.

    Some steps in the execution of a composite target need to execute "actions lite" or anonymous targets. Things that
    have an execute method but are only executable through other targets. This class can be derived to create that
    without generating all the normal target metadata.
    """

    def __init__(self, scope: TargetScope):
        """Set scope of this action"""
        self.scope = scope

    def is_supported(self, builder: "Build", context: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        # Always supported by targets, unless specifically overridden
        return True

    @abstractmethod
    def execute(
        self, builder: "Build", context: Path, args: Tuple[Dict[str, str], List[str]]
    ):
        """Executes the given target"""

    def allows_pass_args(self):
        """Target allows pass-through arguments"""
        return False

    def pass_handler(self):
        """Handler of pass-through args"""
        return None

    def __repr__(self):
        """Representation"""
        return f"{self.__class__.__name__}"


class Target(ExecutableAction):
    """Generic build target base class

    A target can be specified by the user using a mnemonic and flags. The mnemonic is the command typed in by the user,
    and the flags allow the user to remember fewer mnemonics by changing the build target using a modifier. Each build
    target is available in certain build types.

    Targets can be global, using the GlobalTarget base class. Global targets don't use contextual information to modify
    the target, but apply to the whole deployment. Note: global targets are also engaged at the deployment level should
    that be the context.

    Targets may also be local. These targets use context information to figure out what to build. This allows for one
    target to represent a class of targets. i.e. build can be used as a local target to build any given sub directory.
    """

    ALL_TARGETS = []

    def __init__(
        self,
        mnemonic: str,
        desc: str,
        scope: TargetScope,
        build_type: BuildType = None,
        flags: set = None,
    ):
        """Constructs a build target and registers it as one of the global targets

        As part of the construction of a Target it is registered as part of the targets available to be run by
        fprime-util. Targets defined as both global and local are wrapped in delegating targets (one for each scope) and
        those delegates are registered. This is for brevity in definition of these targets. The flag "--all" is added to
        global targets to distinguish them

        Args:
            mnemonic:    mnemonic used to engage build targets. Is not unique, but mnemonic + flags must be.
            desc:        help description of this build target
            build_types: supported build types for target. Defaults to [BuildType.BUILD_NORMAL, BuildType.BUILD_TESTING]
            flags:       flags used to uniquely identify build targets who share logical mnemonics. Defaults to None.
            cmake:       cmake target override to handle oddly named cmake targets
        """
        super().__init__(scope)
        self.mnemonic = mnemonic
        self.desc = desc
        self.build_type = (
            build_type if build_type is not None else BuildType.BUILD_NORMAL
        )
        self.flags = flags if flags is not None else set()

        # Targets defined as either local or global scope are registered directly. "Both" targets are wrapped in a
        # delegator for both scopes and those end up being registered.
        if self.scope != TargetScope.BOTH:
            self.ALL_TARGETS.append(
                self
            )  # Add newly minted target to the tracked list of targets
        else:
            DelegatorTarget(self, mnemonic, desc, TargetScope.LOCAL, build_type, flags)
            new_flags = {"all"}
            new_flags = new_flags.union(flags) if flags else new_flags
            DelegatorTarget(
                self, mnemonic, desc, TargetScope.GLOBAL, build_type, new_flags
            )

    def __repr__(self):
        """Representation"""
        return f"{self.__class__.__name__}({str(self)})"

    def __str__(self):
        """Makes this target into a string"""
        return self.config_string(self.mnemonic, self.flags)

    @staticmethod
    def config_string(mnemonic, flags):
        """Converts a mnemonic and set of flags to string

        Args:
            mnemonic: mnemonic of the target
            flags: set of flags to pair with mnemonic
        Returns:
            string of format "mnemonic --flag1 --flag2 ..."
        """
        flag_string = " ".join([f"--{flag}" for flag in flags])
        flag_string = f" {flag_string}" if flag_string else ""
        return f"{mnemonic}{flag_string}"

    @classmethod
    def get_all_possible_flags(cls) -> Set[str]:
        """Gets list of all targets' flags used

        Returns:
            List of targets supported by the system
        """
        return functools.reduce(
            lambda agg, item: agg.union(item.flags), cls.get_all_targets(), set()
        )

    @classmethod
    def get_all_targets(cls) -> List["Target"]:
        """Gets list of all targets registered

        Returns:
            List of targets supported by the system
        """
        return cls.ALL_TARGETS

    @classmethod
    def get_target(cls, mnemonic: str, flags: Set[str]) -> "Target":
        """Gets the actual build target given the parsed namespace

        Using the global list of build targets and the flags supplied to the namespace, attempt to determine which build
        targets can be used. If more than one are found, then generate exception.

        Args:
            mnemonic: mnemonic of command to look for
            flags:    flags to narrow down target

        Returns:
            single matching target
        """
        matching = [
            target
            for target in cls.get_all_targets()
            if target.mnemonic == mnemonic and flags == target.flags
        ]
        if not matching:
            raise NoSuchTargetException(
                f"Could not find target '{cls.config_string(mnemonic, flags)}'"
            )
        assert len(matching) == 1, "Conflicting targets specified in code"
        return matching[0]


class CompositeTarget(Target):
    """Target whose execution is a composition of other targets"""

    def __init__(self, targets, *args, **kwargs):
        """Constructor setting child targets"""
        super().__init__(*args, **kwargs)
        self.targets = targets

    def __repr__(self):
        """So we can see what it delegated to"""
        return f"{self.__class__.__name__}[{', '.join([target.__repr__() for target in self.targets])}]"

    def is_supported(self, builder: "Build", context: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        # Supported only if all steps supported
        return functools.reduce(
            lambda sum, target: sum and target.is_supported(builder, context),
            self.targets,
            True,
        )

    def allows_pass_args(self):
        """Pass args allowed if any child allows it"""
        return functools.reduce(
            lambda sum, target: sum or target.allows_pass_args(), self.targets, False
        )

    def pass_handler(self):
        """Pass handler as , separated list"""
        handlers = [
            target.pass_handler() for target in self.targets if target.pass_handler()
        ]
        return ",".join(handlers)

    def execute(self, *args, **kwargs):
        """Execute the composite target"""
        for child in self.targets:
            # Composite actions must override scope as a delegator may have acted to change the scope
            old_scope = child.scope
            try:
                child.scope = self.scope
                child.execute(*args, **kwargs)
            finally:
                child.scope = old_scope


class BuildSystemTarget(Target):
    """Target whose execution invokes a command within the build system"""

    def __init__(self, build_target, *args, **kwargs):
        """Constructor setting child targets"""
        super().__init__(*args, **kwargs)
        self.build_target = build_target

    def execute(
        self, builder: "Build", context: Path, args: Tuple[Dict[str, str], List[str]]
    ):
        """Execute a build target

        Executes a target within the build system. This will execute the target by calling into the build system.
        Context is supplied such that the system can match local targets to the global target list.

        Args:
            builder: builder to execute target with
            target: target to run
            context: context path for local targets
            make_args: make system arguments directly supplied
        """
        # Global targets with build target "" must be mapped to "arg"
        build_target = (
            self.build_target
            if self.build_target != "" or self.scope == TargetScope.LOCAL
            else "all"
        )
        builder.execute_build_target(
            build_target, context, self.scope == TargetScope.GLOBAL, args[0]
        )

    def is_supported(self, builder: "Build", context: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        build_target_names = builder.cmake.get_available_targets(
            str(builder.build_dir), context
        )
        return self.build_target in build_target_names


class DelegatorTarget(Target):
    """Delegates to another target

    Sometimes a target needs to be created that delegates to another target. As an example, local and global variants of
    a "both" target need to delegate through the original. This target delegates to other targets.
    """

    def __init__(self, delegate: Target, *args, **kwargs):
        """Constructor"""
        super().__init__(*args, **kwargs)
        self.delegate = delegate

    def __repr__(self):
        """So we can see what it delegated to"""
        return f"{self.__class__.__name__}[{self.delegate.__repr__()}]"

    def is_supported(self, builder: "Build", context: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        return self.delegate.is_supported(builder, context)

    def allows_pass_args(self):
        """Pass args allowed if any child allows it"""
        return self.delegate.allows_pass_args()

    def pass_handler(self):
        """Pass handler from delegate"""
        return self.delegate.pass_handler()

    def execute(self, *args, **kwargs):
        """Delegate the execution"""
        old_scope = self.delegate.scope
        try:
            # Temporarily overrides effective scope of delegate for this invocation
            self.delegate.scope = self.scope
            return_value = self.delegate.execute(*args, **kwargs)
        finally:
            self.delegate.scope = old_scope
        return return_value



"""
Created on Dec 18, 2014
@author: tcanham, reder
"""
import struct

from .type_base import ValueType
from .type_exceptions import (
    DeserializeException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
)


class BoolType(ValueType):
    """
    Representation of a boolean type that will be stored for F prime. True values are stored as a U8 of 0xFF and False
    is stored as a U8 of 0x00.
    """

    TRUE = 0xFF
    FALSE = 0x00

    @classmethod
    def validate(cls, val):
        """Validate the given class"""
        if not isinstance(val, bool):
            raise TypeMismatchException(bool, type(val))

    def serialize(self):
        """Serialize a boolean value"""
        if self._val is None:
            raise NotInitializedException(type(self))
        return struct.pack("B", self.TRUE if self._val else self.FALSE)

    def deserialize(self, data, offset):
        """Deserialize boolean value"""
        try:
            int_val = struct.unpack_from("B", data, offset)[0]
            if int_val not in [self.TRUE, self.FALSE]:
                raise TypeRangeException(int_val)
            self._val = int_val == self.TRUE
        except struct.error:
            raise DeserializeException("Not enough bytes to deserialize bool.")

    def getSize(self):
        return struct.calcsize("B")



""" fprime.fbuild.gcovr: coverage target implementations """
import atexit
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Tuple, Union

from .target import ExecutableAction, Target, TargetScope, CompositeTarget

TEMPORARY_DIRECTORY = "{{AUTOCODE}}"


def _get_project_path(builder: "Build", module: Union[str, Path]) -> Path:
    """Get the project path from a given module string

    Get the project path for a given module or path. Paths are assumed to be contexts and returned verbatim. Modules are
    converted to a project string and then that is returned.

    Args:
        builder: builder object
        module: module string or project path
    """
    # If this module is path then return it as project path already
    if isinstance(module, Path):
        return module
    # Calculate the project path from the module name
    project_root = builder.get_settings(
        "project_root", builder.get_settings("framework_path", builder.build_dir.parent)
    )
    return Path(project_root) / module.replace("_", "/")


def _using_root(builder: "Build", context: Path, scope: TargetScope):
    """Are we using repository root for things"""
    return builder.is_deployment(context) or scope == TargetScope.GLOBAL


def _get_ac_directory(builder: "Build", context: Path, scope: TargetScope) -> Path:
    """Gets a directory for AC files to be placed

    AC files need to migrate out of the build cache and into the build for purposes of HTML coverage.  This selects an
    appropriate location for this temporary directory.

    Args:
        builder: build to use
        context: path context to use
    """
    base_path = (
        _get_project_path(builder, ".")
        if _using_root(builder, context, scope)
        else context
    )
    return (base_path / TEMPORARY_DIRECTORY).absolute()


class GcovClean(ExecutableAction):
    """Target used to scrub gcov files before a coverage run

    The way fprime builds unittests means it is easy to see gcov coverage files leak into the coverage report from other
    runs. This target finds all gcov created files (.gcda) and removes them before the build. This applies to the
    entire build cache as report should be complete and isolated to any given run.
    """

    REMOVAL_EXTENSIONS = [".gcda"]

    def execute(
        self, builder: "Build", context: Path, args: Tuple[Dict[str, str], List[str]]
    ):
        """Execute the clean using os.walk"""
        print(f"[INFO] Scrubbing leftover { ', '.join(self.REMOVAL_EXTENSIONS)} files")
        self._recurse(builder.build_dir)

    @classmethod
    def _recurse(cls, parent_path: Path):
        """Recursive helper"""
        for path in parent_path.iterdir():
            if path.is_file() and path.suffix in cls.REMOVAL_EXTENSIONS:
                path.unlink()
            elif path.is_dir():
                cls._recurse(path)


class GcovrAcHelper(ExecutableAction):
    """Helps with AC file paths and gcovr reports

    Autocoded files are created in the build cache, but not in the same directory as the .gcda and .gcno files. This
    causes the --html-details flag to fail to find these files when generating the detailed HTML w/ source output. This
    action will fix this by coping these files to a known working directory to work within.

    Specifically:
    - Create temporary directory
    - Register an at_exit handler to clean-up the directory
    - For each .gcda file in build cache:
    -   Copy matching .cpp and .hpp file from two parent directories above to temporary directory
    """

    EXTENSION_TRIGGER = [".gcno", ".gcda"]

    def __init__(self, scope: TargetScope):
        """Init function"""
        super().__init__(scope)
        self.collided_files = set()

    def execute(
        self, builder: "Build", context: Path, args: Tuple[Dict[str, str], List[str]]
    ):
        """Execute this action"""
        temp_path = _get_ac_directory(builder, context, self.scope)
        print(f"[INFO] Making temporary directory: {temp_path}")
        temp_path.mkdir(exist_ok=False)
        atexit.register(self.removal, temp_path)
        print("[INFO] Copying AC files into temporary directory")
        cache_path = builder.build_dir / builder.get_build_cache_path(context)
        recurse_path = (
            Path(builder.build_dir)
            if _using_root(builder, context, self.scope)
            else cache_path
        )
        # Recurse excluding the autocoders directory which is filled with test codes
        self._recurse(
            recurse_path,
            temp_path,
            [(builder.build_dir / "F-Prime" / "Autocoders").absolute()],
        )
        for item in list(self.collided_files):
            print(f"[WARNING] Removing contested file: {item.name}")
            item.unlink()

    def copy_safe(self, source: Path, destination: Path):
        """Copy one file if it doesn't exist. Checking content if it does.

        Copy if the destination does not exist. If the content differs, add the offending file to a list to clean up
        later such that it cannot explode gcovr. Deferring removal prevents a third copy from winning.

        Args:
            source: source file
            destination: destination file
        """
        if not destination.exists():
            shutil.copy(source, destination)
            return
        with open(source) as file1:
            lines1 = file1.readlines()
        with open(destination) as file2:
            lines2 = file2.readlines()
        # Line number deltas are known to cause problems as gcovr becomes unhappy when the line number becomes
        # inconsistent with the source it is reading.
        if len(lines1) != len(lines2):
            self.collided_files.add(destination)
            return
        for line1, line2 in zip(lines1, lines2):
            # Lines that that match and lines that differ only by the include statement are ok. These are the models
            # that implement multiple physical components.
            if (line1.strip() == line2.strip()) or (
                line1.strip().startswith("#include")
                and line2.strip().startswith("#include")
            ):
                continue
            # Real differences count as collisions
            self.collided_files.add(destination)

    @staticmethod
    def removal(directory: Path):
        """Function to remove temporary directory"""
        print(f"[INFO] Removing temporary directory: {directory}")
        shutil.rmtree(directory)

    def copy_gcda_pair(self, destination: Path, gcda: Path):
        """Copy matching CPP and HPP files given an .gcda"""
        base_name = gcda.name
        for extension in self.EXTENSION_TRIGGER:
            base_name = base_name.replace(extension, "")
        ac_folder = gcda.parent.parent.parent

        # Copy both possible AC files into the destination folder
        for possible in [base_name, base_name.replace(".cpp", ".hpp")]:
            source = ac_folder / possible
            if source.exists():
                self.copy_safe(source, destination / possible)

    def _recurse(self, parent_path: Path, destination: Path, excludes: List[Path]):
        """Recursive helper"""
        for path in parent_path.iterdir():
            if path.absolute() in excludes:
                continue
            if path.is_file() and path.suffix in self.EXTENSION_TRIGGER:
                self.copy_gcda_pair(destination, path)
            elif path.is_file() and path.name.endswith("Ac.hpp"):
                self.copy_safe(path, destination / path.name)
            elif path.is_dir():
                self._recurse(path, destination, excludes)


class Gcovr(ExecutableAction):
    """Target invoking `gcovr` utility to calculate coverage

    `gcovr` is a utility used to calculate coverage recursively. It is a wrapper around the gcov executable and patches
    several problems with the utility. This target runs `gcovr` and places the output in a newly created coverage
    directory. arguments are supplied into the `gcovr` execution, however; the default execution is always supplied the
    --txt coverage/summary.txt --txt stdout flags.
    """

    EXECUTABLE = "gcovr"

    def is_supported(self, builder: "Build", context: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        return bool(shutil.which(self.EXECUTABLE))

    def execute(
        self, builder: "Build", context: Path, args: Tuple[Dict[str, str], List[str]]
    ):
        """Executes the gcovr target"""
        if not shutil.which(self.EXECUTABLE):
            print(
                f"[ERROR] Cannot find executable: {self.EXECUTABLE}. Unable to run coverage report.",
                file=sys.stderr,
            )
            return
        ac_temporary_path = _get_ac_directory(builder, context, self.scope)
        coverage_output_dir = context / "coverage"
        coverage_output_dir.mkdir(exist_ok=True)
        project_root = builder.get_settings(
            "project_root",
            builder.get_settings("framework_path", builder.build_dir.parent.parent),
        )
        filter_path = (
            Path(project_root).absolute()
            if _using_root(builder, context, self.scope)
            else _get_project_path(builder, context)
        )
        build_offset = (
            ""
            if _using_root(builder, context, self.scope)
            else builder.get_build_cache_path(context)
        )

        # gcovr is an unhappy beast (but it is *much* better than the raw gcov executable
        # In order to deal with gcovr, we need to do the following things:
        #   1. Setup the project root (deals with relative file paths within project). This is set to our project_root.
        #   2. Next setup the search path for object files. This is set to the build cache or build module directory.
        #   3. Specify two filters: project_path (module directory) and autocode temp directory
        #   4. Exclude the build cache, and autocoders
        #   5. Then supply default output options
        #   6. Then supply pass through arguments
        cli_args = [
            "gcovr",
            "-r",
            str(project_root),
            builder.build_dir / build_offset,  # For efficiency in searching on modules
            "--filter",
            f"{filter_path}/*",
            "--filter",
            f"{ac_temporary_path}",
            "--exclude",
            str(builder.build_dir),
            "--exclude",
            str(
                builder.get_settings("framework_path", builder.build_dir.parent.parent)
                / "Autocoders"
            ),
            "--print-summary",
            "--txt",
            f"{coverage_output_dir}/summary.txt",
            f"--html-details",
            f"{coverage_output_dir}/coverage{'-all' if self.scope == TargetScope.GLOBAL else ''}.html",
        ]
        cli_args.extend(args[1])

        if builder.cmake.verbose:
            print(f"[INFO] Running '{' '.join(cli_args)}'")
        # gcovr must run in the ac_temporary_path or html details cannot find the Ac files
        subprocess.call(cli_args, cwd=str(ac_temporary_path))

    def allows_pass_args(self):
        """Gcovr allows pass-through arguments"""
        return True

    def pass_handler(self):
        """Pass handler"""
        return self.EXECUTABLE


class GcovrTarget(CompositeTarget):
    """Target specific to gcovr

    The gcovr target is a composite target that builds upon an existing check target to perform gcovr work. In addition
    it must support extra arguments as we pass these to the gcovr executable.
    """

    def __init__(self, check_target: Target, scope: TargetScope, *args, **kwargs):
        """Construct the gcovr target around an existing check target"""
        assert check_target.scope in [
            scope,
            TargetScope.BOTH,
        ], "Cannot create composite target from incompatible target"
        super().__init__(
            [GcovClean(scope), check_target, GcovrAcHelper(scope), Gcovr(scope)],
            scope=scope,
            *args,
            **kwargs,
        )



from __future__ import print_function

import datetime
import os
from os.path import join
from fprime.fbuild.settings import IniSettings
from fprime.fbuild.builder import Build
from pathlib import Path
import textwrap
from fprime.fbuild.interaction import replace_contents


def remove_line(filename, removal):
    with open(filename, "r") as f:
        lines = f.readlines()
        f.close()
    with open(filename, "w") as f:
        for line in lines:
            if line != removal:
                f.write(line)
        f.close()


def update_sdd(component_kind, commands, parameters, events, telemetry):
    ports = "## Port Descriptions\n| Name | Description |\n"
    if component_kind == "active":
        ports += textwrap.dedent(
            """\
        | PingIn | Used for pinging other components |
        | PingOut | Used to receive ping signal |\n"""
        )
    elif component_kind == "queued":
        ports += ""
        ports += textwrap.dedent(
            """\
            | PingIn | Used for pinging other components |
            | PingOut | Used to receive ping signal |
            | SchedIn | Used as a schedular for queued components |\n"""
        )

    if commands == "yes":
        replace_contents(
            "docs/sdd.md",
            "## Commands",
            textwrap.dedent(
                """\
        ## Commands
        | Name | Description |
        | ExampleCommand | Example of how a command is implemented |\n"""
            ),
        )
        ports += textwrap.dedent(
            """\
            | cmdIn | Used to receive commands |
            | cmdRegOut | Used to register with the command dispatcher |
            | cmdResponseOut | Used to return responses to command dispatcher |\n"""
        )
    else:
        remove_line("docs/sdd.md", "## Commands\n")

    if parameters == "yes":
        replace_contents(
            "docs/sdd.md",
            "## Parameters",
            textwrap.dedent(
                """\
        ## Parameters
        | Name | Description |
        | ExampleParameter | Example of how a parameter is implemented |\n"""
            ),
        )
        ports += textwrap.dedent(
            """\
            | prmGetOut | Used to get parameter value |
            | prmSetOut | Used to set parameter value |\n"""
        )
    else:
        remove_line("docs/sdd.md", "## Parameters\n")

    if events == "yes":
        replace_contents(
            "docs/sdd.md",
            "## Events",
            textwrap.dedent(
                """\
        ## Events
        | Name | Description |
        | EX_ExampleEvent | Example of how an event is implemented |\n"""
            ),
        )
        ports = ports + textwrap.dedent(
            """\
            | eventOut | Used to send events |
            | textEventOut | Used to send text formatted events |\n"""
        )
    else:
        remove_line("docs/sdd.md", "## Events\n")

    if telemetry == "yes":
        replace_contents(
            "docs/sdd.md",
            "## Telemetry",
            textwrap.dedent(
                """\
        ## Telemetry
        | Name | Description |
        | ExampleChannel | Example of how a telemetry channel is implemented |\n"""
            ),
        )
        ports = ports + textwrap.dedent(
            """\
            | tlmOut | Used to send telemetry data |\n"""
        )
    else:
        remove_line("docs/sdd.md", "## Telemetry\n")

    ports = ports + textwrap.dedent(
        """\
            | timeGetOut | Used to pass time stamps around the system |\n"""
    )
    replace_contents("docs/sdd.md", "## Port Descriptions", ports)


def main():
    cwd = Path(os.getcwd())
    deployment = Build.find_nearest_deployment(cwd)
    settings = IniSettings.load(Path(deployment, "settings.ini"))
    proj_root_found = settings.get("project_root") is not None
    today = datetime.date.today()
    replace_contents(join("docs", "sdd.md"), "<TODAY>", today.strftime("%m/%d/%Y"))
    update_sdd(
        "{{ cookiecutter.component_kind }}",
        "{{ cookiecutter.commands }}",
        "{{ cookiecutter.parameters }}",
        "{{ cookiecutter.events }}",
        "{{ cookiecutter.telemetry }}",
    )
    remove_line("{{cookiecutter.component_name}}ComponentAi.xml", "\n")

    print(
        """
    ################################################################################
    ################################################################################

        You have successfully created the `{{ cookiecutter.component_name }}` component.

    ################################################################################

        You've used these cookiecutter parameters:
    {% for key, value in cookiecutter.items()|sort %}
            {{ "{0:26}".format(key + ":") }} {{ "{0!r}".format(value).strip("u") }}
    {%- endfor %}

    ################################################################################

        Now you can edit your {{ cookiecutter.component_name }}Ai.xml file
        define the component to your liking

        In addition, a sdd.md file has been created in the docs directory 
        for you to document your component
    """
    )

    if not proj_root_found:
        print(
            """
        No project root was specified in your settings.ini file.
        This means you will need to add this component to your build
        system and then possibly purge and generate your project.

        In addition, the unit test files were not generated.
            
        To fix this issue, ensure you have a settings.ini file that 
        specifies your project_root
            """
        )


if __name__ == "__main__":
    main()



"""
Created on Dec 18, 2014

@author: tcanham

"""
import struct

from fprime.constants import DATA_ENCODING

from . import type_base
from .type_exceptions import (
    DeserializeException,
    NotInitializedException,
    StringSizeException,
    TypeMismatchException,
)


class StringType(type_base.DictionaryType):
    """
    String type representation for F prime. This is a type that stores a half-word first for representing the length of
    this given string. Each sub string class defines the sub-type property max length that represents the maximum
    length of any string value stored to it.

    All string types follow this implementation, but have some specific type-based properties: MAX_LENGTH.
    """

    @classmethod
    def construct_type(cls, name, max_length=None):
        """Constructs a new string type with given name and maximum length"""
        return type_base.DictionaryType.construct_type(cls, name, MAX_LENGTH=max_length)

    @classmethod
    def validate(cls, val):
        """Validates that this is a string"""
        if not isinstance(val, str):
            raise TypeMismatchException(str, type(val))
        if cls.MAX_LENGTH is not None and len(val) > cls.MAX_LENGTH:
            raise StringSizeException(len(val), cls.MAX_LENGTH)

    def serialize(self):
        """
        Serializes the string in a binary format
        """
        # If val is never set then it is init exception...
        if self.val is None:
            raise NotInitializedException(type(self))
        # Check string size before serializing
        elif self.MAX_LENGTH is not None and len(self.val) > self.MAX_LENGTH:
            raise StringSizeException(len(self.val), self.MAX_LENGTH)
        # Pack the string size first then return the encoded data buffer
        return struct.pack(">H", len(self.val)) + self.val.encode(DATA_ENCODING)

    def deserialize(self, data, offset):
        """
        Deserializes a string from the given data buffer.
        """
        try:
            val_size = struct.unpack_from(">H", data, offset)[0]
            # Deal with not enough data left in the buffer
            if len(data[offset + 2 :]) < val_size:
                raise DeserializeException(
                    f"Not enough data to deserialize string data. Needed: {val_size} Left: {len(data[offset + 2 :])}"
                )
            # Deal with a string that is larger than max string
            if self.MAX_LENGTH is not None and val_size > self.MAX_LENGTH:
                raise StringSizeException(val_size, self.MAX_LENGTH)
            self.val = data[offset + 2 : offset + 2 + val_size].decode(DATA_ENCODING)
        except struct.error:
            raise DeserializeException("Not enough bytes to deserialize string length.")

    def getSize(self):
        """
        Get the size of this object
        """
        return struct.calcsize(">H") + len(self.val)



"""
numerical_types.py:

A file that contains the definitions for all the integer types provided as part of F prime.  F prime supports integers
that map to stdint.h integer sizes, that is, 8-bit, 16-bit, 32-bit, and 64-bit signed and unsigned integers.

@author mstarch
"""
import abc
import struct

from .type_base import ValueType
from .type_exceptions import (
    DeserializeException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
)


class NumericalType(ValueType, abc.ABC):
    """Numerical types that can be serialized using struct and are of some power of 2 byte width"""

    @classmethod
    @abc.abstractmethod
    def get_bits(cls):
        """Gets the integer bits of a given type"""

    @classmethod
    def getSize(cls):
        """Gets the size of the integer based on the size specified in the class name"""
        return int(cls.get_bits() >> 3)  # Divide by 8 quickly

    @staticmethod
    @abc.abstractmethod
    def get_serialize_format():
        """Gets the format serialization string such that the class can be serialized via struct"""
        raise NotImplementedError("get_serialize_format")

    def serialize(self):
        """Serializes this type using struct and the val property"""
        if self._val is None:
            raise NotInitializedException(type(self))
        return struct.pack(self.get_serialize_format(), self._val)

    def deserialize(self, data, offset):
        """Serializes this type using struct and the val property"""
        try:
            self._val = struct.unpack_from(self.get_serialize_format(), data, offset)[0]
        except struct.error as err:
            raise DeserializeException(str(err))


class IntegerType(NumericalType, abc.ABC):
    """Base class that represents all integer common functions"""

    @classmethod
    @abc.abstractmethod
    def range(cls):
        """Gets signed/unsigned of this type"""

    @classmethod
    def validate(cls, val):
        """Validates the given integer."""
        if not isinstance(val, int):
            raise TypeMismatchException(int, type(val))
        min_val, max_val = cls.range()
        if val < min_val or val > max_val:
            raise TypeRangeException(val)


class FloatType(NumericalType, abc.ABC):
    """Base class that represents all float common functions"""

    @classmethod
    def validate(cls, val):
        """Validates the given integer."""
        if not isinstance(val, (float, int)):
            raise TypeMismatchException(float, type(val))


class I8Type(IntegerType):
    """Single byte integer type. Represents C chars"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (-128, 127)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 8

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return "b"


class I16Type(IntegerType):
    """Double byte integer type. Represents C shorts"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (-32768, 32767)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 16

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">h"


class I32Type(IntegerType):
    """Four byte integer type. Represents C int32_t,"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (-2147483648, 2147483647)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 32

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">i"


class I64Type(IntegerType):
    """Eight byte integer type. Represents C int64_t,"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (-9223372036854775808, 9223372036854775807)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 64

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">q"


class U8Type(IntegerType):
    """Single byte integer type. Represents C chars"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (0, 0xFF)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 8

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return "B"


class U16Type(IntegerType):
    """Double byte integer type. Represents C shorts"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (0, 0xFFFF)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 16

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">H"


class U32Type(IntegerType):
    """Four byte integer type. Represents C unt32_t,"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (0, 0xFFFFFFFF)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 32

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">I"


class U64Type(IntegerType):
    """Eight byte integer type. Represents C unt64_t,"""

    @classmethod
    def range(cls):
        """Gets signed/unsigned of this type"""
        return (0, 0xFFFFFFFFFFFFFFFF)

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 64

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">Q"


class F32Type(FloatType):
    """Eight byte integer type. Represents C unt64_t,"""

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 32

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">f"


class F64Type(FloatType):
    """Eight byte integer type. Represents C unt64_t,"""

    @classmethod
    def get_bits(cls):
        """Get the bit count of this type"""
        return 64

    @staticmethod
    def get_serialize_format():
        """Allows serialization using struct"""
        return ">d"



"""
fprime.fbuild.cmake:

File to contain basic wrappers for the CMake. This will enable basic CMake commands in order to detect the properties of
the build. This should not be imported in-person, but rather should be included by the build package. This can be the
receiver of these delegated functions.

@author mstarch
"""
# Get a cache directory for building CMakeList file, if need and remove at exit
import collections
import copy
import functools
import itertools
import os
import pty
import re
import selectors
import shutil
import subprocess
import sys
import time

from fprime.common.error import FprimeException


class CMakeHandler:
    """
    CMake handler interacts with an F prime CMake-based system. This will help us interact with CMake in refined ways.
    """

    CMAKE_LOCATION_FIELDS = [
        "FPRIME_PROJECT_ROOT",
        "FPRIME_LIBRARY_LOCATIONS",
        "FPRIME_FRAMEWORK_PATH",
    ]

    def __init__(self):
        """Instantiate a basic CMake handler"""
        self.settings = {}
        self._cmake_cache = None
        self.verbose = False
        self.cached_help_targets = None
        try:
            self._run_cmake(["--help"], print_output=False)
        except Exception as exc:
            raise CMakeExecutionException(
                "CMake executable 'cmake' not found", str(exc), printed=False
            ) from exc

    def set_verbose(self, verbose):
        """Sets verbosity"""
        self.verbose = verbose

    def validate_cmake_cache(self, cmake_args, build_dir):
        cmake_cache = self._read_cache(build_dir)
        for key, expected in cmake_args.items():
            actual = cmake_cache.get(key, None)
            # Allow expected variables not to be set in cache.
            # When building with a newer version of fprime-util and an older version of F prime,
            # newer CMake options that aren't used by older versions of CMake won't be set in the cache
            if actual is None:
                continue
            expected = str(expected).strip()
            actual = str(actual).strip()
            if expected != actual:
                raise CMakeInconsistentCacheException(key, expected, actual)

    def execute_known_target(
        self,
        target,
        build_dir,
        path,
        cmake_args=None,
        make_args=None,
        top_target=False,
        environment=None,
        full_cache_rebuild=False,
        print_output=True,
    ):
        """
        Executes a known target for a given build_dir. Path will default to a known path.

        :param build_dir: build_dir to use to run this.
        :param target: target to execute at the path, using above build_dir
        :param path: path to run target against. (default) current working directory
        :param cmake_args: cmake args input
        :param top_target: top-level target. Do not append path name
        :param environment: environment to setup when executing CMake
        :param full_cache_rebuild: rebuild the cache fully. Default: False, use the short version
        :param print_output: should this call print output. Default: True, print output
        :return: return code from CMake
        """
        assert build_dir is not None, "Invalid build dir supplied"
        build_dir = str(build_dir)
        cmake_args = {} if cmake_args is None else cmake_args
        self.validate_cmake_cache(cmake_args, build_dir)

        make_args = {} if make_args is None else make_args
        fleshed_args = ["--"] + list(
            map(lambda key: f"{key}={make_args[key]}", make_args.keys())
        )

        if target != "" and top_target:
            cmake_target = target
        else:
            module = self.get_cmake_module(path, build_dir)
            cmake_target = (
                module
                if target == ""
                else (f"{module}_{target}".lstrip("_") if not top_target else target)
            )
        run_args = ["--build", build_dir]
        environment = {} if environment is None else copy.copy(environment)
        if self.verbose:
            environment["VERBOSE"] = "1"

        # The cmake target is "" when running build in the root of the F' repo.
        if cmake_target != "":
            run_args.extend(["--target", cmake_target])

        try:
            return self._run_cmake(
                run_args + fleshed_args,
                write_override=True,
                environment=environment,
                print_output=print_output,
            )
        except CMakeExecutionException as exc:
            no_target = functools.reduce(
                lambda cur, line: cur or "No rule to make target" in line,
                exc.get_errors(),
                False,
            )
            if not no_target:
                raise
            print(
                "[CMAKE] CMake failed to detect target, attempting CMake cache refresh and retry"
            )
            self.cmake_refresh_cache(build_dir, full=full_cache_rebuild)
            return self._run_cmake(
                run_args + fleshed_args,
                write_override=True,
                environment=environment,
                print_output=print_output,
            )

    def get_include_locations(self, cmake_dir):
        """
        Gets the locations that can be used as the root of an include tree. Common directories are placed in these
        include locations. These include standard builds, configs, etc.

        :param cmake_dir: directory of a CMake build, or directory containing a CMake project
        :return: []  List of include locations. Order: project, lib, lib, ..., F prime core
        """
        # Reading config fields. If the cmake_dir is a project dir, a build cache may be setup.
        # !! Note: using a project dir will cause file-system side effects, and incur a one-time cost to setup cache !!
        config_fields = self.get_fprime_configuration(
            CMakeHandler.CMAKE_LOCATION_FIELDS, cmake_dir
        )
        non_null = filter(lambda item: item is not None and item != "", config_fields)
        # Read cache fields for each possible directory the build_dir, and the new tempdir
        locations = itertools.chain(*map(lambda value: value.split(";"), non_null))
        mapped = map(os.path.abspath, locations)
        # Removes duplicates, by creating an ordered dictionary, and then asking for its keys
        return list(collections.OrderedDict.fromkeys(mapped).keys())

    def get_include_info(self, path, cmake_dir):
        """
        Calculates the include root of the given path. The include root is defined as the following based on the other
        two values supplied. First, the following paths values are established:
        - Location of the project's root. This is defined in the project_dir's CMakeList.txt, or in the CMake Cache.
        - Location of the project's F prime checkout. This is defined in the same places.
        - Locations of any libraries defined in the same places.
        From there, the include root of the supplied path is whichever of those paths is your parent. In cases where
        both are parents, it will take the nearest parent.

        :param path: path to calculate looking for include-root
        :param cmake_dir: directory of a CMake build, or directory containing a CMake project
        :return: (relative include path, include root for the given path)
        """
        path = (
            os.path.abspath(path) if path is not None else os.path.abspath(os.getcwd())
        ) + os.sep
        possible_parents = self.get_include_locations(cmake_dir)
        # Check there is some possible parent
        if not possible_parents:
            raise CMakeProjectException(
                cmake_dir,
                "Does not define any of the cache fields: {}".format(
                    ",".join(CMakeHandler.CMAKE_LOCATION_FIELDS)
                ),
            )
        # Common-prefix will fail in the case that two directories share a common prefix in their name
        # So force directory separator to force common prefix to only work on full directory names
        full_parents = map(lambda dir: os.path.abspath(dir) + os.sep, possible_parents)
        # Parents *are* the common prefix for their children
        parents = filter(
            lambda parent: os.path.commonprefix([parent, path]) == parent, full_parents
        )

        def parent_reducer(accum, item):
            """
            Reduces a list of parents, and the given path, to a single path who is the closes parent to the existing
            path while stilling being a parent.

            :param accum: latest best parent, or None if no valid parent found
            :param item: current parent being evaluated
            :return: next latest best parent
            """
            # None items cannot weigh-in
            if item is None:
                return accum
            # Calculate common path
            prefix = os.path.commonprefix([item, path])
            # Return the previous parent, if new prefix is not the directory itself, or the old one was closer
            return (
                accum
                if prefix != item or (accum is not None and len(accum) > len(item))
                else item
            )

        parents = list(parents)
        nearest_parent = functools.reduce(parent_reducer, parents, None)
        # Check that a parent is the true parent
        if nearest_parent is None:
            raise CMakeOrphanException(path)
        return os.path.relpath(path, nearest_parent), nearest_parent.rstrip(os.sep)

    def get_fprime_configuration(self, fields, cmake_dir=None):
        """
        Gets fprime configuration for the given field(s). This will return a list of fields for the set of input fields.
        The user may supply a string for a single value returned as list of one, or a list of fields for list of values.

        :param fields: name of field, or list of names of fields
        :param cmake_dir: a cmake directory (project or build) to used. default: None, use existing temp cached build.
        :return: list of values, or Nones
        """
        if isinstance(fields, str):
            fields = [fields]
        # Setup the build_dir if it can be detected. Without a cache or specified value, we can crash
        self._cmake_validate_build_dir(cmake_dir)  # Validate the dir
        return self._read_values_from_cache(fields, build_dir=cmake_dir)

    def generate_build(
        self, source_dir, build_dir, args=None, ignore_output=False, environment=None
    ):
        """Generate a build directory for purposes of the build.

        :param source_dir: source directory to generate from
        :param build_dir: build directory to generate to
        :param args: arguments to hand to CMake.
        :param ignore_output: do not print the output where the user can see it
        :param environment: environment to set when generating
        """

        if not os.path.exists(build_dir):
            os.makedirs(build_dir)
        # We will CD for build, so this path must become absolute
        source_dir = os.path.abspath(source_dir)
        args = {} if args is None else args
        fleshed_args = map(
            lambda key: ("{}={}" if key.startswith("--") else "-D{}={}").format(
                key, args[key]
            ),
            args.keys(),
        )
        self.cmake_validate_source_dir(source_dir)
        self._run_cmake(
            ["-S", source_dir] + list(fleshed_args),
            workdir=build_dir,
            print_output=not ignore_output,
            write_override=True,
            environment=environment,
        )

    def get_cmake_module(self, path, build_dir):
        """Gets the CMake module

        CMake modules are constructed from a path relative to some project root (fprime, deployment, or library roots).
        This relative path is then converted to use "_" instead of "/"

        Args:
            path: path to contextualize. May be None to use os.getcwd().
            build_dir: build directory to use

        Returns:
            CMake module name in format x_y_z
        """
        project_relative_path = self.get_project_relative_path(path, build_dir)
        module = project_relative_path.replace(
            ".", ""
        )  # Handles case where relative path is exactly "."
        module = module.replace(os.sep, "_")
        return module

    def get_project_relative_path(self, path, build_dir):
        """Gets the path relative to the cmake setup, or raises CMakeOrphanException

        Args:
            path: path to contextualize. May be None to use os.getcwd().
            build_dir: build directory to use

        Returns:
            path string that is relative to some root of the project. i.e. used for target suffix names
        """
        # Get module name from the relative path to include root
        include_root = self.get_include_info(path, build_dir)[1]
        return os.path.relpath(path, include_root)

    def get_available_targets(self, build_dir, path):
        """Gets a list of available CMake targets in the current directory

        Args:
            build_dir: build directory to use for detecting targets
            path: contextual path. None for "cwd"

        Note:
            This code might not work on non-GNU makefile variants of CMake as it depends on the "help" target

        Returns:
            list of contextual CMake make targets
        """
        if not self.cached_help_targets:
            run_args = ["--build", build_dir, "--target", "help"]
            stdout, _ = self._run_cmake(
                run_args, write_override=True, print_output=False
            )
            self.cached_help_targets = [
                line.replace("...", "").strip()
                for line in stdout
                if line.startswith("...")
            ]

        prefix = self.get_cmake_module(path, build_dir)
        return [
            make.replace(prefix, "").strip("_")
            for make in self.cached_help_targets
            if make.startswith(prefix)
        ]

    @staticmethod
    def purge(build_dir):
        """
        Reusable purge functionality, so the user may purge or the system may cleanup itself

        :param build_dir: build dir specified to purge
        """
        shutil.rmtree(build_dir, ignore_errors=True)

    def _read_values_from_cache(self, keys, build_dir):
        """
        Reads set values from cache into an output tuple.

        :param keys: keys to read in iterable
        :param build_dir: build directory containing cache file
        :return: a tuple of keys, None if not part of cache
        """
        cache = self._read_cache(build_dir)
        # Reads cache values suppressing KeyError, {}.get(x, default=None)
        miner = lambda x: cache.get(x, None)
        return tuple(map(miner, keys))

    def _read_cache(self, build_dir):
        """
        Reads the cache from the associated build_dir. This will return a dictionary of cache variable name to
        its value. The first run will call cmake to extract the cache and cache it internally. Subsequent
        calls will return the cached cache.

        :param build_dir: build directory to harvest for cache variables
        :return: {<cmake cache variable>: <cmake cache value>}
        """
        if self._cmake_cache is not None:
            return self._cmake_cache

        reg = re.compile("([^:]+):[^=]*=(.*)")
        # Check that the build_dir is properly setup
        self._cmake_validate_build_dir(build_dir)
        stdout, _ = self._run_cmake(["-LA"], workdir=build_dir, print_output=False)
        # Scan for lines in stdout that have non-None matches for the above regular expression
        valid_matches = filter(lambda item: item is not None, map(reg.match, stdout))
        # Return the dictionary composed from the match groups
        self._cmake_cache = dict(
            map(lambda match: (match.group(1), match.group(2)), valid_matches)
        )
        return self._cmake_cache

    @staticmethod
    def cmake_validate_source_dir(source_dir):
        """
        Raises an exception if the source dir is not a valid CMake source directory. This means a CMakeLists.txt exists
        and defines within it a project call.

        :param source_dir: source directory to validate
        """
        cmake_file = os.path.join(source_dir, "CMakeLists.txt")
        if not os.path.isfile(cmake_file):
            raise CMakeProjectException(source_dir, "No CMakeLists.txt is defined")
        # Test the cmake_file for project(
        with open(cmake_file, encoding="utf8") as file_handle:
            project_lines = list(
                filter(lambda line: "project(" in line, file_handle.readlines())
            )
            if not project_lines:
                raise CMakeProjectException(
                    source_dir, f"No 'project()' calls in {cmake_file}"
                )

    @staticmethod
    def _cmake_validate_build_dir(build_dir):
        """
        Raises an exception if the build dir is not a valid CMake build directory

        :param build_dir: build_dir to validate
        """
        cache_file = os.path.join(build_dir, "CMakeCache.txt")
        if not os.path.isfile(cache_file):
            raise CMakeInvalidBuildException(build_dir)

    def cmake_refresh_cache(self, build_dir, full=False):
        """
        Runs the cmake  target required to refresh the cmake cache. This will allow for unknown targets to be searched
        for before the utility gives up and produces.

        :param build_dir: directory to build in
        :param full: full re-generate of the cache. Default: false, attempt to build 'noop' target instead
        """
        if full:
            environment = {}
            run_args = ["--build", build_dir]
            if self.verbose:
                print("[CMAKE] Refreshing CMake build cache")
                environment["VERBOSE"] = "1"
            run_args.extend(["--target", "rebuild_cache"])
            self._cmake_cache = None  # Clear internal cache of cmake variables
            self._run_cmake(
                run_args,
                write_override=True,
                environment=environment,
                print_output=self.verbose,
            )
        else:
            if self.verbose:
                print("[CMAKE] Checking CMake cache for rebuild")
            self.execute_known_target(
                "noop",
                build_dir,
                None,
                top_target=True,
                full_cache_rebuild=True,
                print_output=self.verbose,
            )

    def _run_cmake(
        self,
        arguments,
        workdir=None,
        print_output=True,
        write_override=False,
        environment=None,
    ):
        """
        Will run the cmake system supplying the given arguments. Assumes that the CMake executable is somewhere on the
        path in order for this to run.

        :param arguments: arguments to supply to CMake.
        :param workdir: work directory to run in
        :param print_output: print_output to the screen. Default: True
        :param write_override: allows for non-read-only commands
        :param environment: environment to write into
        :return: (stdout, stderr)
        """
        if environment is None:
            environment = {}

        cm_environ = os.environ.copy()
        cm_environ.update(self.settings.get("environment", {}))
        cm_environ.update(environment)
        cargs = ["cmake"]
        if not write_override:
            cargs.append("-N")
        cargs.extend(arguments)
        if self.verbose:
            print(f"""[CMAKE] '{" ".join(cargs)}'""")
            for key, val in sorted(cm_environ.items(), key=lambda x: x[0]):
                print(f"[CMAKE]     {key}={val}")

        # In order to get proper console highlighting while still getting access to the output, we need to create a
        # pseudo-terminal. This will allow the proc to write to one side, and our select below to read from the other
        # side. Most importantly, pseudo-terminal usage will trick CMake into highlighting for us.
        pty_err_r, pty_err_w = pty.openpty()
        pty_out_r, pty_out_w = pty.openpty()
        proc = subprocess.Popen(
            cargs,
            stdout=pty_out_w,
            stderr=pty_err_w,
            cwd=workdir,
            env=cm_environ,
            close_fds=True,
        )
        # Close our side of the writing pipes. Subproc will also close when it it finished.
        os.close(pty_out_w)
        os.close(pty_err_w)
        ret, stdout, stderr = self._communicate(
            proc, open(pty_out_r, mode="rb"), open(pty_err_r, mode="rb"), print_output
        )
        # Raising an exception when the return code is non-zero allows us to handle the exception internally if it is
        # needed. Thus we do not just exit.
        if ret != 0:
            raise CMakeExecutionException(
                f"CMake erred with return code {proc.returncode}",
                stderr,
                print_output,
                ret,
            )
        return stdout, stderr

    @staticmethod
    def _communicate(proc, stdout, stderr, print_output=True):
        """
        Communicates with a process while assuring that output is captured and optionally printed. This will buffer
        lines for the standard out file handle when not none, and will always buffer standard error so that it can be
        provided when needed. This effectively replaces the .communicate method of the proc itself, while still
        preventing deadlocks.  The output is returned for each stream as a list of lines.

        :param proc: Popen process constructed with the above pairs to the submitted file handles
        :param stdout: standard output file handle. Paired with the FH provided to the Popen stdout argument
        :param stderr: standard error file handle. Paired with the FH provided to the Popen stderr argument
        :param print_output: print output to the screen. If False, output is masked. Default: True, print to screen.
        :return return code, stdout as list of lines, stderr as list of lines
        """
        stdouts = []
        stderrs = []
        # Selection will allow us to read from stdout and stderr whenever either is available. This will allow the
        # program to capture both, while still printing as soon as is possible. This will keep the user informed, not
        # merge streams, and capture output.
        #
        # Selection blocks on the read of "either" file descriptor, and then passes off the execution to the below code
        # with a key that represents which descriptor was the one available to read without blocking.
        selector = selectors.DefaultSelector()
        selector.register(stdout, selectors.EVENT_READ, data=(stdouts, sys.stdout))
        selector.register(stderr, selectors.EVENT_READ, data=(stderrs, sys.stderr))
        while not stdout.closed or not stderr.closed:
            # This line *BLOCKS* until on of the above registered handles is available to read. Then a set of events is
            # returned signaling that a given object is available for IO.
            events = selector.select()
            for key, _ in events:
                appendable, stream = key.data
                try:
                    line = key.fileobj.readline().decode().replace("\r\n", "\n")
                # Some systems (like running inside Docker) raise an io error instead of returning "" when the device
                # is ended. Not sure why this is, but the effect is the same, on IOError assume end-of-input
                except OSError:
                    line = ""
                appendable.append(line)
                # Streams are EOF when the line returned is empty. Once this occurs, we are responsible for closing the
                # stream and thus closing the select loop. Empty strings need not be printed.
                if not line:
                    key.fileobj.close()
                    continue
                # Forwards output to screen.  Assuming a PTY is used, then coloring highlights should be automatically
                # included for output. Raw streams are used to avoid print quirks
                if print_output:
                    stream.write(line)
                    stream.flush()
        # Spin waiting for the .poll() method to return a non-None result ensuring that the process has finished.
        while proc.poll() is None:
            time.sleep(0.0001)
        return proc.poll(), stdouts, stderrs


class CMakeException(FprimeException):
    """Error occurred within this CMake package"""

    def __init__(self, message, exit_code=1):
        super().__init__(message)
        self.exit_code = exit_code


class CMakeInconsistentCacheException(CMakeException):
    """Project CMakeCache.txt is inconsistent with settings.ini file"""

    def __init__(self, key, expected, actual):
        """Force an appropriate message"""
        super().__init__(
            f"Expected CMake variable {key} to be set to '{expected}', was actually set to '{actual}'. This is usually caused by updating the settings.ini file without purging and regenerating the accompanying build cache."
        )


class CMakeOrphanException(CMakeException):
    """File is not managed by CMake project"""

    def __init__(self, file_dir):
        """Force an appropriate message"""
        super().__init__(f"{file_dir} is outside the F prime project")


class CMakeProjectException(CMakeException):
    """Invalid project directory"""

    def __init__(self, project_dir, error):
        """Force an appropriate message"""
        super().__init__(f"{project_dir} is an invalid F prime deployment. {error}")


class CMakeInvalidBuildException(CMakeException):
    """The supplied build directory was not setup as a CMake value"""

    def __init__(self, build_dir):
        """Force an appropriate message"""
        super().__init__(
            f"{build_dir} is not a CMake build directory. Please setup using 'fprime-util generate'"
        )


class CMakeExecutionException(CMakeException):
    """Pass up a CMake Error as an Exception"""

    def __init__(self, message, stderr, printed, exit_code=1):
        """The error data should be stored"""
        super().__init__(message, exit_code)
        self.stderr = stderr
        self.need = not printed

    def get_errors(self):
        """
        Returns the error stream data to the caller
        :return: stderr of CMake as supplied into this Exception
        """
        return self.stderr

    def need_print(self):
        """
        Returns if the errors need to be printed
        :return: need print
        """
        return self.need


class CMakeNoSuchTargetException(CMakeException):
    """The target does not exist."""

    def __init__(self, build_dir, target):
        """Better messaging for this exception"""
        super().__init__(f"{build_dir} does not support target {target}")
